
from logging import getLogger

from .database import add_enum_constant, add_statistics
from ..pipeline.calculate.power import PowerModel, PowerCalculator, BikeModel

log = getLogger(__name__)

POWER_MODEL_CNAME = 'power-model'


def add_simple_power_estimate(s, c, activity_group, cda, crr, bike_weight, rider_weight):
    '''
    Configure parameters directly.
    '''
    power_model = add_enum_constant(s, POWER_MODEL_CNAME, PowerModel,
                                    {'bike_model': {'cda': cda, 'crr': crr, 'bike_weight': bike_weight},
                                     'rider_weight': rider_weight},
                                    single=False, activity_group=activity_group, description='''
Parameters used in estimating power (for the given activity group).

This is the direct (simpler) configuration (unless it has been modified).  
The physical parameters are encoded directly below.
* CdA is the coefficient of drag multiplied by frontal area.
* Crr is the coefficient of rolling resistance.
* bike_weight is the weight of the bike.
* rider_weight is the weight of the rider.
In more complex configurations these can be references to other constants.
''')
    add_statistics(s, PowerCalculator, c, power_model=power_model.name, activity_group=activity_group.name)


def add_kit_power_estimate(s, c, activity_groups):
    '''
    Configure parameters indirectly.  This uses a single constant, but must be run for each group.
    The constant doesn't specify the group, so the current group will be used.
    '''
    power_model = add_enum_constant(s, POWER_MODEL_CNAME, PowerModel,
                                    # no activity group is given so the default on expansion will be used
                                    # which will be taken from the context (in this case the activity)
                                    {'bike_model': '${Constant.power-${SegmentReader.kit}}',
                                     'rider_weight': '${DiaryTopic.Weight:}'},
                                    single=True, description='''
Parameters used in estimating power (for the given activity group).

This is the indirect (complex) configuration which delegates to a kit-specific bike model and reads
the weight from a diary entry.
''')
    for activity_group in activity_groups:
        add_statistics(s, PowerCalculator, c, power_model=power_model.name, activity_group=activity_group)


def add_kit_power_model(s, kit, activity_group, cda, crr, bike_weight):
    add_enum_constant(s, 'power-' + kit, BikeModel,
                      {'cda': cda, 'crr': crr, 'bike_weight': bike_weight},
                      single=False, activity_group=activity_group, description='''
Parameters used in estimating power (for the given kit and activity group).
* CdA is the coefficient of drag multiplied by frontal area.
* Crr is the coefficient of rolling resistance.
* bike_weight is the weight of the bike.
''')



from logging import getLogger

from .lib.log import make_log_from_args, set_log_color

log = getLogger(__name__)

__GLOBAL_DEV = False
__GLOBAL_DATA = None


def set_global_dev(dev):
    global __GLOBAL_DEV
    __GLOBAL_DEV = dev
    log.debug(f'Setting global dev flag: {dev}')


def global_dev():
    global __GLOBAL_DEV
    return __GLOBAL_DEV


def set_global_data(data):
    global __GLOBAL_DATA
    __GLOBAL_DATA = data


def global_data():
    global __GLOBAL_DATA
    return __GLOBAL_DATA


def set_global_state(args):
    from .sql.system import Data
    from .commands.args import DEV, BASE
    set_global_dev(args[DEV])
    make_log_from_args(args)
    data = Data(args[BASE])
    set_global_data(data)
    set_log_color(args, data.sys)
    return data




from enum import IntEnum
from json import dumps
from logging import getLogger

from sqlalchemy import Column, Integer, not_, or_

from ..support import Base
from ..types import Cls, Json, Sort

log = getLogger(__name__)


class PipelineType(IntEnum):

    CALCULATE = 0
    DISPLAY = 1
    DISPLAY_ACTIVITY = 4
    READ_ACTIVITY = 2
    READ_MONITOR = 3


class Pipeline(Base):

    __tablename__ = 'pipeline'

    id = Column(Integer, primary_key=True)
    type = Column(Integer, nullable=False, index=True)
    cls = Column(Cls, nullable=False)
    args = Column(Json, nullable=False, server_default=dumps(()))
    kargs = Column(Json, nullable=False, server_default=dumps({}))
    sort = Column(Sort, nullable=False)

    @classmethod
    def _query(cls, s, type=None, like=tuple(), unlike=tuple(), id=None):
        q = s.query(Pipeline)
        if type is not None:  # enum can be 0
            q = q.filter(Pipeline.type == type)
        if like:
            q = q.filter(or_(*[Pipeline.cls.like(pattern) for pattern in like]))
        if unlike:
            q = q.filter(not_(or_(*[Pipeline.cls.like(pattern) for pattern in unlike])))
        if id:
            q = q.filter(Pipeline.id == id)
        return q

    @classmethod
    def all(cls, s, type, like=tuple(), unlike=tuple(), id=None):
        q = cls._query(s, type, like=like, unlike=unlike, id=id)
        pipelines = q.order_by(Pipeline.sort).all()
        if not pipelines:
            msg = 'No pipelines configured for type %s' % PipelineType(type).name
            if like:
                msg += f' like {like}'
            if id:
                msg += f' with ID={id}'
            raise Exception(msg)
        yield from pipelines
        
    @classmethod
    def all_instances(cls, s, type, like=tuple(), unlike=tuple(), id=None):
        for pipeline in cls.all(s, type, like=like, unlike=unlike, id=id):
            log.debug(f'Building {pipeline.cls} ({pipeline.args}, {pipeline.kargs})')
            yield pipeline.cls(*pipeline.args, **pipeline.kargs)

    @classmethod
    def count(cls, s, type, like=tuple(), unlike=tuple(), id=None):
        q = cls._query(s, type, like=like, unlike=unlike, id=id)
        return q.count()



from logging import getLogger
from tempfile import TemporaryDirectory

from ch2.commands.read import read
from ch2.commands.args import bootstrap_dir, mm, DEV, V, m
from ch2.config.profile.default import default
from ch2.diary.database import read_date
from ch2.diary.model import LABEL, VALUE
from ch2.lib import to_date
from tests import LogTestCase

log = getLogger(__name__)


class TestModel(LogTestCase):

    def test_constant(self):
        with TemporaryDirectory() as f:
            bootstrap_dir(f, m(V), '5', mm(DEV), configurator=default)
            args, data = bootstrap_dir(f, m(V), '5', mm(DEV),
                                       'read', 'data/test/source/personal/2018-03-04-qdp.fit',
                                       '-Kn_cpu=1')
            read(args, data)
            with data.db.session_context() as s:
                model = list(read_date(s, to_date('2018-03-04')))
                for i, x in enumerate(model):
                    print(i, x)
                [title, diary, shrimp, activity, database] = model
                activity = activity[1][2]  # multiple now supported
                print(activity)
                name = activity[1]
                print(name)
                self.assertEqual(name[LABEL], 'Name')
                self.assertEqual(name[VALUE], '2018-03-04T07:16:33')
                route = activity[2]
                self.assertEqual(route[LABEL], 'Route')



import datetime as dt
from logging import getLogger
from re import search

from sqlalchemy import or_, distinct, asc, desc
from sqlalchemy.orm.exc import NoResultFound

from ..utils import Displayer, ActivityJournalDelegate
from ...calculate import SummaryCalculator
from ...calculate.activity import ActivityCalculator
from ...calculate.power import PowerCalculator
from ....data.climb import climbs_for_activity
from ....diary.database import summary_column, interval_column
from ....diary.model import optional_text, text, from_field, value
from ....lib import local_date_to_time, time_to_local_time, to_time, to_date, time_to_local_date, \
    log_current_exception
from ....lib.date import YMD_HM, HM, format_minutes, add_date, MONTH, YMD, YEAR, YM, DAY
from ....names import Names as N
from ....sql import ActivityGroup, ActivityJournal, ActivityTopicJournal, ActivityTopicField, StatisticName, \
    ActivityTopic, StatisticJournal, Pipeline, PipelineType, Interval

log = getLogger(__name__)


class ActivityDisplayer(Displayer):

    @optional_text('Activities')
    def _read_date(self, s, date):
        start = local_date_to_time(date)
        finish = start + dt.timedelta(days=1)
        for activity_group in s.query(ActivityGroup).order_by(ActivityGroup.sort).all():
            for ajournal in s.query(ActivityJournal). \
                    filter(ActivityJournal.finish >= start,
                           ActivityJournal.start < finish,
                           ActivityJournal.activity_group == activity_group). \
                    order_by(ActivityJournal.start).all():
                yield list(self._single_activity(s, ajournal, date))

    def _single_activity(self, s, ajournal, date):
        yield text(self.__title(s, ajournal), tag='activity-title', db=ajournal)
        yield from self._read_journal_date(s, ajournal, date)

    @staticmethod
    def __title(s, ajournal):
        return '%s - %s  (%s) %s' % (time_to_local_time(to_time(ajournal.start), fmt=YMD_HM),
                                     time_to_local_time(to_time(ajournal.finish), fmt=HM),
                                     format_minutes((ajournal.finish - ajournal.start).seconds),
                                     ajournal.activity_group.name)

    def _read_journal_date(self, s, ajournal, date):
        for pipeline in Pipeline.all_instances(s, PipelineType.DISPLAY_ACTIVITY):
            try:
                if pipeline.interpolate:
                    yield from pipeline.read_journal_date(s, ajournal, date)
                else:
                    entry = list(pipeline.read_journal_date(s, ajournal, date))
                    if entry: yield entry
            except Exception as e:
                log.warning(f'Error calling {pipeline}')
                log_current_exception(traceback=True)

    @optional_text('Activities')
    def _read_schedule(self, s, date, schedule):
        intervals = s.query(Interval).join(ActivityGroup). \
            filter(Interval.schedule == schedule,
                   Interval.start == date).all()
        intervals = sorted(intervals, key=lambda i: i.activity_group.name if i.activity_group else '')
        for interval in intervals:
            pipelines = list(self._read_pipelines(s, interval))
            if pipelines:
                yield [text(interval.activity_group.name, tag='activity-group')] + pipelines

    def _read_pipelines(self, s, interval):
        for pipeline in Pipeline.all_instances(s, PipelineType.DISPLAY_ACTIVITY):
            try:
                 results = list(pipeline.read_interval(s, interval))
                 if results: yield results
            except Exception as e:
                log.warning(f'Error calling {pipeline} with {interval}')
                log_current_exception(traceback=True)


class ActivityDelegate(ActivityJournalDelegate):

    def __init__(self, *args, **kargs):
        super().__init__(interpolate=True, *args, **kargs)

    def read_journal_date(self, s, ajournal, date):
        yield list(self._read_journal_topics(s, ajournal, date))
        yield from self.__read_details(s, ajournal, date)

    def _read_journal_topics(self, s, ajournal, date):
        yield text('Activity', db=ajournal)
        yield from self.__read_activity_topics(s, ajournal, date)

    def __read_activity_topics(self, s, ajournal, date):
        tjournal = ActivityTopicJournal.get_or_add(s, ajournal.file_hash, ajournal.activity_group)
        cache = tjournal.cache(s)
        # special case root
        for field in s.query(ActivityTopicField). \
                join(ActivityTopic). \
                filter(ActivityTopic.title == ActivityTopic.ROOT,
                       or_(ActivityTopic.activity_group_id == None,
                           ActivityTopic.activity_group_id == ajournal.activity_group.id)). \
                order_by(ActivityTopicField.sort).all():
            yield from_field(field, cache[field])
        for topic in s.query(ActivityTopic). \
                filter(ActivityTopic.parent == None,
                       ActivityTopic.title != ActivityTopic.ROOT,
                       or_(ActivityTopic.activity_group_id == None,
                           ActivityTopic.activity_group_id == ajournal.activity_group.id)). \
                order_by(ActivityTopic.sort).all():
            yield list(self.__read_activity_topic(s, date, cache, topic))

    def __read_activity_topic(self, s, date, cache, topic):
        yield text(topic.name)
        if topic.description: yield text(topic.description)
        log.debug(f'topic id {topic.id}; fields {topic.fields}')
        for field in topic.fields:
            yield from_field(field, cache[field])
        for child in topic.children:
            if child.activity_group == topic.activity_group and child.schedule.at_location(date):
                content = list(self.__read_activity_topic(s, date, cache, child))
                if content: yield content

    def __read_details(self, s, ajournal, date):
        zones = list(self.__read_zones(s, ajournal))
        if zones: yield [text('HR Zones (% time)')] + zones
        active_data = list(self.__read_active_data(s, ajournal, date))
        if active_data: yield [text('Activity Statistics')] + active_data
        climbs = list(self.__read_climbs(s, ajournal, date))
        if climbs: yield [text('Climbs')] + climbs
        for (title, template, re) in (('Min Time', N.MIN_KM_TIME_ANY, r'(\d+km)'),
                                      ('Med Time', N.MED_KM_TIME_ANY, r'(\d+km)'),
                                      ('Max Med Heart Rate', N.MAX_MED_HR_M_ANY, r'(\d+m)'),
                                      ('Max Mean Power Estimate', N.MAX_MEAN_PE_M_ANY, r'(\d+m)')):
            model = list(self.__read_template(s, ajournal, template, re, date))
            if model: yield [text(title)] + model

    @staticmethod
    def __read_zones(s, ajournal):
        percent_times = s.query(StatisticJournal).join(StatisticName). \
            filter(StatisticJournal.time == ajournal.start,
                   StatisticName.name.like(N.PERCENT_IN_Z_ANY),
                   StatisticName.owner == ActivityCalculator) \
            .order_by(StatisticName.name).all()
        if percent_times:
            for zone, percent_time in reversed(list(enumerate((time.value for time in percent_times), start=1))):
                yield [value('Zone', zone, tag='hr-zone'), value('Percent time', percent_time)]

    @staticmethod
    def __read_active_data(s, ajournal, date):
        for name in (N.ACTIVE_DISTANCE, N.ACTIVE_TIME, N.ACTIVE_SPEED, N.MEAN_POWER_ESTIMATE):
            sjournal = StatisticJournal.at(s, ajournal.start, name, ActivityCalculator, ajournal.activity_group)
            if sjournal:
                yield value(sjournal.statistic_name.title, sjournal.value,
                            units=sjournal.statistic_name.units,
                            measures=sjournal.measures_as_model(date))
        for name in (N._delta(N.FITNESS_ANY), N._delta(N.FATIGUE_ANY),
                     N.EARNED_D_ANY, N.RECOVERY_D_ANY):
            for sjournal in StatisticJournal.at_like(s, ajournal.start, name, ActivityCalculator,
                                                     ajournal.activity_group):
                yield value(sjournal.statistic_name.title, sjournal.value,
                            units=sjournal.statistic_name.units,
                            measures=sjournal.measures_as_model(date))
        for name in (N.ENERGY_ESTIMATE, N.CALORIE_ESTIMATE):
            sjournal = StatisticJournal.at(s, ajournal.start, name, PowerCalculator, ajournal.activity_group)
            if sjournal:
                yield value(sjournal.statistic_name.title, sjournal.value,
                            units=sjournal.statistic_name.units,
                            measures=sjournal.measures_as_model(date))

    @classmethod
    def __sjournal_as_value(cls, sjournal, date=None):
        measures = sjournal.measures_as_model(date) if date else None
        return value(sjournal.statistic_name.title, sjournal.value,
                     units=sjournal.statistic_name.units, measures=measures)

    @classmethod
    def __climb_as_value(cls, climb, key, date=None):
        return cls.__sjournal_as_value(climb[key], date=date)

    @classmethod
    def __read_climbs(cls, s, ajournal, date):
        total, climbs = climbs_for_activity(s, ajournal)
        if total:
            yield cls.__sjournal_as_value(total, date=date)
            for climb in climbs:
                yield [text('Climb'),
                       cls.__climb_as_value(climb, N.CLIMB_ELEVATION, date=date),
                       cls.__climb_as_value(climb, N.CLIMB_DISTANCE),
                       cls.__climb_as_value(climb, N.CLIMB_TIME),
                       cls.__climb_as_value(climb, N.CLIMB_GRADIENT)]

    def __read_template(self, s, ajournal, template, re, date):
        sjournals = s.query(StatisticJournal).join(StatisticName). \
            filter(StatisticJournal.time == ajournal.start,
                   StatisticName.name.like(template),
                   StatisticName.owner == ActivityCalculator).order_by(StatisticName.name).all()
        for sjournal in self.__sort_journals(sjournals):
            if sjournal.value > 0:  # avoid zero power and anything else with silly value
                yield value(search(re, sjournal.statistic_name.title).group(1), sjournal.value,
                            units=sjournal.statistic_name.units, measures=sjournal.measures_as_model(date))

    @staticmethod
    def __sort_journals(sjournals):
        return sorted(sjournals,
                      # order by integer embedded in name
                      key=lambda sjournal: int(search(r'(\d+)', sjournal.statistic_name.title).group(1)))

    @staticmethod
    def __sort_names(names):
        # order by integer embedded in name
        return sorted(names, key=lambda name: int(search(r'(\d+)', name).group(1)))


    @optional_text('Activities', tag='activity')
    def read_interval(self, s, interval):
        for name in (N.ACTIVE_DISTANCE, N.ACTIVE_TIME, N.ACTIVE_SPEED, N.TOTAL_CLIMB, N.CLIMB_ELEVATION,
                     N.CLIMB_DISTANCE, N.CLIMB_GRADIENT, N.CLIMB_TIME):
            column = list(interval_column(s, interval, name, SummaryCalculator))
            if column: yield column
        for template in (N.MIN_KM_TIME_ANY, N.MED_KM_TIME_ANY, N.MAX_MED_HR_M_ANY):
            for name in self.__sort_names(self.__names_like(s, template)):
                column = list(interval_column(s, interval, name, SummaryCalculator))
                if column: yield column

    @staticmethod
    def __names_like(s, name):
        return [statistic_name.name
                for statistic_name in s.query(StatisticName).
                    filter(StatisticName.name.like(name),
                           StatisticName.owner == ActivityCalculator).all()]


def active_dates(s, start, range, fmt):
    date_start = to_date(start)
    date_end = add_date(date_start, (1, range))
    time_start = local_date_to_time(date_start)
    time_end = local_date_to_time(date_end)
    times = s.query(distinct(ActivityJournal.start)). \
        filter(ActivityJournal.start >= time_start,
               ActivityJournal.start < time_end).all()
    return list(set(time_to_local_date(row[0]).strftime(fmt) for row in times))


def active_days(s, month):
    return active_dates(s, month, MONTH, YMD)


def active_months(s, year):
    return active_dates(s, year, YEAR, YM)


def latest_activity(s):
    return s.query(ActivityJournal). \
        order_by(desc(ActivityJournal.start)).limit(1).one_or_none()


def activity_times_by_group(s):
    by_group = {}
    for group in s.query(ActivityGroup).all():
        activities = [time_to_local_time(activity.start)
                      for activity in s.query(ActivityJournal).
                          filter(ActivityJournal.activity_group == group).
                          order_by(desc(ActivityJournal.start)).all()]
        if activities:
            by_group[group.name] = activities
    return by_group


from logging import getLogger

from .args import QUERY, SUB_COMMAND, ACTIVITIES, SHOW, SET, mm
from ..data.constraint import activity_conversion, constrained_sources, sort_groups, \
    group_by_type
from ..diary.model import TEXT

log = getLogger(__name__)


def search(args, data):
    '''
## search

    > ch2 search text QUERY [--show NAME ...] [--set NAME=VALUE]
    > ch2 search activities QUERY [--show NAME ...] [--set NAME=VALUE]
    > ch2 search sources QUERY [--show NAME ...] [--set NAME=VALUE]

Search the database.

The first form (search text) searches for the given text in activity name and description.

The second form (search activities) is similar, but allows for more complex searches (similar to SQL)
that target particular fields.

The third form (search sources) looks for matches for any source (not just activities).

Note that 'search activities' treats both activity journals and activity topics (ie data from FIT
files and data entered by the user) as a single 'source', while 'search activities' treats each source
as separate.

Once a result is found additional statistics from that source be displayed (--show)
and a single value modified (--set).

The search syntax (for activities and sources) is similar to SQL, but element names are statistic names.
A name has the format "Owner.name:group" where the owner and group are optional.
A trailing colon implies a NULL group (used for statistics that are not specific to any activity, like diary entries).
The name and group also include SQL wildcards (eg "%fitness%").

The owner of a name is the process that calculated the value.
It works as a kind of "namespace" - the database could contain multiple statistics called "active_distance"
but only one will have been calculated by ActivityCalculator.

In addition, attributes of the source can be accessed using "Class.attribute" where Class is optional.
For showing or setting values on the result, Class must be omitted (so .start=... sets the start attribute).

For complex searches, string values must be quoted, negation and NULL values are not supported,
and comparison must be between a name and a value (not two names).

There is experimental support for null values (actually missing values).
The form of the query is less general than SQL - a field must always be compared with a value (not another field).

### Examples

    > ch2 search text bournemouth

Find any activities where the text mentions Bournemouth.

    > ch2 search sources 'name="Wrong Name"' --set 'name="Right Name"'

Modify the name variable.

    > ch2 search activities 'ActivityCalculator.active_distance:mtb > 10 and active_time < 3600'

Find mtb activities that cover over 10km in under an hour.

    > ch2 search activities 'name="%"' --show .start name

Find activities that have a defined name and display both the name and the activity start time
(the 'dot' syntax allows access to an attribute on the found activity).

    >  ch2 search activities 'ActivityJournal.start=2020-04-17T09:27:30' --set name='Corral Quemado'

Set the name on the activity at the given time (again, using the dot syntax).
    '''
    if args[SHOW] and args[SET]:
        raise Exception(f'Give at most one of {mm(SHOW)} and {mm(SET)}')
    cmd = args[SUB_COMMAND]
    with data.db.session_context() as s:
        if cmd == TEXT:
            query = ' and '.join([f'(ActivityTopic.name = "{word}" or ActivityTopic.notes = "{word}")'
                                  for word in args[QUERY]])
            conversion = activity_conversion
        else:
            query = ' '.join(args[QUERY])
            if cmd == ACTIVITIES:
                conversion = activity_conversion
            else:
                conversion = None
        results = constrained_sources(s, query, conversion=conversion)
        process_results(s, results, show=args[SHOW], set=args[SET], activity=bool(conversion))


def text_search(s, words):
    query = ' and '.join([f'(ActivityTopic.name = "{word}" or ActivityTopic.notes = "{word}")'
                          for word in words])
    return constrained_sources(s, query, activity_conversion)


def process_results(s, sources, show=None, set=None, activity=False):
    groups = sort_groups(group_by_type(sources))
    for type in groups:
        print(f'\n{type.__name__}:')
        if show:
            show_results(s, groups[type], show, activity)
        elif set:
            set_results(s, groups[type], set, activity)
        else:
            for source in groups[type]:
                print('  ', source.long_str())
    if show:
        print()


def show_results(s, sources, show, activity):
    for i, source in enumerate(sources, start=1):
        for qname in show:
            if qname.startswith('.'):
                cls = source.__class__.__name__
                try:
                    value = getattr(source, qname[1:])
                    print(f'{i:03d}  {cls}{qname} {value}')
                except AttributeError:
                    print(f'{i:03d}  {cls}{qname} not present')
            else:
                if activity:
                    journals = source.get_all_qname(s, qname)
                else:
                    journals = source.get_qname(s, qname)
                for journal in journals:
                    warning = 'TIME SERIES ' if journal.serial is not None else ''
                    owner = journal.statistic_name.owner
                    name = journal.statistic_name.name
                    group = journal.source.activity_group.name if journal.source.activity_group else 'none'
                    print(f'{i:03d}  {warning}{owner}.{name}:{group} {journal.value}')


def set_results(s, sources, set, activity):
    name, value = set.split('=')
    name = name.strip()
    value = value.strip()
    for source in sources:
        if name.startswith('.'):
            instance, attr = source, name[1:]
        else:
            if activity:
                instance, attr = source.get_all_qname(s, name), 'value'
            else:
                instance, attr = source.get_qname(s, name), 'value'
            if not instance: raise Exception(f'{name} not found')
            if len(instance) > 1: raise Exception(f'{name} ambiguous')
            instance = instance[0]
        print(f'{instance}.{attr} <- {value}')
        current = getattr(instance, attr)
        if isinstance(current, int): value = int(value)
        elif isinstance(current, float): value = float(value)
        else: value = drop_quotes(value)
        # don't convert times - let the type handle conversion
        setattr(instance, attr, value)


def drop_quotes(value):
    if len(value) > 1 and value[0] == '"' and value[-1] == '"': value = value[1:-1]
    elif len(value) > 1 and value[0] == "'" and value[-1] == "'": value = value[1:-1]
    return value




from logging import getLogger
from os.path import exists

from matplotlib import use
from matplotlib.pyplot import show, figure

from .args import ACTIVITY, base_system_path, THUMBNAIL, BASE
from ..data.query import Statistics
from ..names import Names
from ..pipeline.read.segment import SegmentReader
from ..sql import ActivityJournal

log = getLogger(__name__)


def thumbnail(args, data):
    '''
## thumbnail

    > ch2 thumbnail ACTIVITY-ID
    > ch2 thumbnail DATE

Generate a thumbnail map of the activity route.
    '''
    with data.db.session_context() as s:
        activity_id = parse_activity(s, args[ACTIVITY])
        # display(s, activity_id)
        create_in_cache(args[BASE], s, activity_id)


def parse_activity(s, text):
    try:
        return int(text)
    except ValueError:
        return ActivityJournal.at(s, text).id


def read_activity(s, activity_id, decimate=10):
    try:
        activity_journal = s.query(ActivityJournal).filter(ActivityJournal.id == activity_id).one()
        df = Statistics(s, activity_journal=activity_journal). \
            by_name(SegmentReader, Names.SPHERICAL_MERCATOR_X, Names.SPHERICAL_MERCATOR_Y).df
        return df.iloc[::decimate, :]
    except:
        raise Exception(f'{activity_id} is not a valid activity ID')


def stats(zs):
    lo, hi = min(zs), max(zs)
    mid = (lo + hi) / 2
    return lo, mid, hi, hi - lo


def normalize(points):
    xs, ys = zip(*points)
    xlo, xmid, xhi, dx = stats(xs)
    ylo, ymid, yhi, dy = stats(ys)
    if dx > dy:
        ylo -= (dx - dy) / 2
    else:
        xlo -= (dy - dx) / 2
    d = max(dx, dy)
    return [(x - xlo) / d - 0.5 for x in xs], [(y - ylo) / d - 0.5 for y in ys], d


def make_figure(xs, ys, side, grid, cm, border):
    fig = figure(frameon=False)
    fig.set_size_inches(cm / 2.54, cm / 2.54)
    ax = fig.add_subplot(1, 1, 1)
    for edge in ('top', 'right', 'bottom', 'left'):
        ax.spines[edge].set_visible(False)
    lim = 0.5 * (1 + border)
    km = 1000 / side
    n = int(lim / (grid * km)) + 1
    ticks = [km * d * grid for d in range(-n, n+1)]
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)
    ax.tick_params(labelbottom=False, labelleft=False, length=0)
    ax.grid(axis='both', color='#535353')
    ax.set_xlim([-lim, lim])
    ax.set_ylim([-lim, lim])
    ax.set_aspect(aspect='equal', adjustable='box')
    ax.plot(xs, ys, color='white')
    ax.plot([xs[0]], [ys[0]], marker='o', color='green', markersize=cm*3)
    ax.plot([xs[-1]], [ys[-1]], marker='o', color='red', markersize=cm*1.5)
    return fig


def fig_from_df(df, grid=10, cm=1.5, border=0.2):
    points = [(x, y) for _, (x, y) in df.iterrows()]
    if points:
        xs, ys, side = normalize(points)
    else:
        xs, ys, side = [0, 0], [0, 0], 1
    return make_figure(xs, ys, side, grid, cm, border)


def display(s, activity_id):
    df = read_activity(s, activity_id)
    use('PyQt5')
    fig = fig_from_df(df)
    show()


def create_in_cache(base, s, activity_id):
    path = base_system_path(base, subdir=THUMBNAIL, file=f'{activity_id}.png')
    if not exists(path):
        df = read_activity(s, activity_id)
        use('agg')
        fig = fig_from_df(df)
        fig.savefig(path, transparent=True)
    log.info(f'Thumbnail in {path}')
    return path



from abc import abstractmethod
from glob import iglob
from logging import getLogger
from os.path import join
from time import time

from ... import FatalException
from ...commands.args import base_system_path, PERMANENT
from ...fit.format.read import filtered_records
from ...lib import to_time, log_current_exception
from ...lib.io import modified_file_scans
from ..pipeline import LoaderMixin, MultiProcPipeline
from ...sql import Timestamp

log = getLogger(__name__)


class AbortImport(Exception):
    pass


class AbortImportButMarkScanned(AbortImport):
    pass


class FitReaderMixin(LoaderMixin):

    def __init__(self, *args, paths=None, sub_dir=None, **kargs):
        self.paths = paths
        self.sub_dir = sub_dir
        super().__init__(*args, **kargs)

    def _delete(self, s):
        # not used
        pass

    def _expand_paths(self, s, paths):
        from ...commands.read import DOT_FIT
        if paths: return paths
        data_dir = base_system_path(self._data.base, version=PERMANENT)
        if self.sub_dir:
            data_dir = join(data_dir, self.sub_dir)
        else:
            log.warning('No sub_dir defined - will scan entire tree')
        return iglob(join(data_dir, '**/*' + DOT_FIT), recursive=True)

    def _missing(self, s):
        return modified_file_scans(s, self._expand_paths(s, self.paths), self.owner_out, self.force)

    def _run_one(self, s, file_scan):
        try:
            self._read(s, file_scan)
            file_scan.last_scan = time()
        except AbortImportButMarkScanned as e:
            log.warning(f'Could not process {file_scan} (scanned)')
            # log_current_exception()
            file_scan.last_scan = time()
        except FatalException:
            raise
        except Exception as e:
            log.warning(f'Could not process {file_scan} (ignored)')
            log_current_exception()

    def _read(self, s, file_scan):
        source, data = self._read_data(s, file_scan)
        s.commit()
        with Timestamp(owner=self.owner_out, source=source).on_success(s):
            loader = self._get_loader(s, add_serial=True)
            self._load_data(s, loader, data)
            loader.load()
        return loader  # returned so coverage can be accessed

    @staticmethod
    def read_fit_file(data, *options):
        types, messages, records = filtered_records(data)
        return [record.as_dict(*options)
                for _, _, record in sorted(records,
                                           key=lambda r: r[2].timestamp if r[2].timestamp else to_time(0.0))]

    @staticmethod
    def _first(path, records, *names):
        return FitReaderMixin.assert_contained(path, records, names, 0)

    @staticmethod
    def _last(path, records, *names):
        return FitReaderMixin.assert_contained(path, records, names, -1)

    @staticmethod
    def assert_contained(path, records, names, index):
        try:
            return [record for record in records if record.name in names][index]
        except IndexError:
            msg = f'No {names} entry(s) in {path}'
            log.debug(msg)
            raise AbortImportButMarkScanned(msg)

    @abstractmethod
    def _read_data(self, s, file_scan):
        raise NotImplementedError()

    @abstractmethod
    def _load_data(self, s, loader, data):
        raise NotImplementedError()


def quote(text):
    return '"' + text + '"'


class MultiProcFitReader(FitReaderMixin, MultiProcPipeline):

    def _args(self, missing, start, finish):
        paths = ' '.join(quote(file_scan.path) for file_scan in missing[start:finish+1])
        log.info(f'Starting worker for {missing[start]} - {missing[finish]}')
        return paths


from logging import getLogger

from ..lib import log_current_exception
from ..config.database import add
from ..sql import Segment

log = getLogger(__name__)


def import_segment(record, old, new):
    if not segment_imported(record, new):
        record.info('Importing segment entries')
        try:
            with old.session_context() as old_s:
                with new.session_context() as new_s:
                    copy_segments(record, old_s, old, new_s)
        except Exception as e:
            log_current_exception()
            record.warning(f'Aborting segment import: {e}')
    else:
        record.warning('Segment entries already imported')


def segment_imported(record, new):
    with new.session_context() as new_s:
        return bool(new_s.query(Segment).count())


def copy_segments(record, old_s, old, new_s):
    segment = old.meta.tables['segment']
    for old_segment in old_s.query(segment).all():
        title = old_segment.name if hasattr(old_segment, 'name') else old_segment.title
        add(new_s, Segment(title=title, description=old_segment.description,
                           distance=old_segment.distance,
                           start_lat=old_segment.start_lat, start_lon=old_segment.start_lon,
                           finish_lat=old_segment.finish_lat, finish_lon=old_segment.finish_lon))
        record.info(f'Segment {title}')




from bokeh.io import output_file

from ch2.data import *
from ch2.data.plot.calendar import *
from ch2.jupyter.decorator import template
from ch2.lib import *
from ch2.names import N
from ch2.pipeline.owners import *


@template
def calendar():

    '''
    # Calendar
    '''

    '''
    $contents
    '''

    s = session('-v2')
    output_file(filename='/dev/null')

    '''
    ## Distance

    Larger distances have larger symbols.
    
    Place the cursor over the symbol for more information.
    '''

    df = Statistics(s). \
        by_name(ActivityTopic, N.NAME). \
        by_name(ActivityCalculator, N.ACTIVE_DISTANCE, N.ACTIVE_TIME, N.TOTAL_CLIMB).with_. \
        copy({N.ACTIVE_DISTANCE: N.ACTIVE_DISTANCE_KM}).add_times().df
    df['Duration'] = df[N.ACTIVE_TIME].map(format_seconds)
    if present(df, N.TOTAL_CLIMB):
        df.loc[df[N.TOTAL_CLIMB].isna(), [N.TOTAL_CLIMB]] = 0

    calendar = Calendar(df, title=N.DISTANCE, not_hover=[N.ACTIVE_DISTANCE, N.ACTIVE_TIME])
    calendar.std_distance()

    '''
    ## Work Done and Fatigue

    Larger increases in Fitness have larger symbols.  Higher fatigue (relative to fitness) is redder.
    
    Place the cursor over the symbol for more information.
    '''

    df = Statistics(s). \
        by_name(ActivityTopic, N.NAME). \
        by_name(ActivityCalculator, N.ACTIVE_DISTANCE, N.ACTIVE_TIME, N.TOTAL_CLIMB). \
        by_name(ActivityCalculator, N._delta(N.FITNESS_ANY), like=True).df
    df = Statistics(s). \
        by_name(ResponseCalculator, N.ANY_FATIGUE_ANY, N.ANY_FITNESS_ANY, like=True). \
        with_.drop_prefix(N.DEFAULT).into(df, tolerance='30m')
    df['Duration'] = df[N.ACTIVE_TIME].map(format_seconds)
    df.loc[df[N.TOTAL_CLIMB].isna(), N.TOTAL_CLIMB] = 0
    work_done = sorted_numeric_labels(df.columns, N._delta(N.FITNESS))[0]
    fitness = sorted_numeric_labels(df.columns, N.FITNESS)[0]
    fatigue = sorted_numeric_labels(df.columns, N.FATIGUE)[0]
    print(fatigue, fitness)
    df['FF Ratio'] = df[fatigue] / df[fitness]

    calendar = Calendar(df, title='Work Done and Fatigue',
                        not_hover=[N.ACTIVE_DISTANCE, N.ACTIVE_TIME] +
                                  [column for column in df.columns if ':' in column])
    calendar.background('square', fill_alpha=0, line_alpha=1, color='lightgrey')
    calendar.set_palette('FF Ratio', K2R, lo=0.5, hi=2)
    calendar.set_size(work_done, min=0.1, gamma=0.5)
    calendar.foreground('square', fill_alpha=1, line_alpha=0)
    calendar.show()

    '''
    ## Distance, Climb and Direction

    Larger distances have larger symbols.  Higher climbs are redder.  
    The arc indicates the general direction relative to the start.
    
    Place the cursor over the symbol for more information.
    '''

    df = Statistics(s). \
        by_name(ActivityTopic, N.NAME). \
        by_name(ActivityCalculator, N.ACTIVE_DISTANCE, N.ACTIVE_TIME, N.TOTAL_CLIMB, N.DIRECTION,
                N.ASPECT_RATIO).df
    df[N.DISTANCE_KM] = df[N.ACTIVE_DISTANCE]
    df['Duration'] = df[N.ACTIVE_TIME].map(format_seconds)
    if present(df, N.TOTAL_CLIMB):
        df.loc[df[N.TOTAL_CLIMB].isna(), N.TOTAL_CLIMB] = 0

    calendar = Calendar(df, title='Distance, Climb and Direction',
                        not_hover=[N.ACTIVE_DISTANCE, N.ACTIVE_TIME] +
                                  [column for column in df.columns if ':' in column])
    calendar.std_distance_climb_direction()

    '''
    ## Distance, Work Done and Direction

    Larger distances have larger symbols.  Larger gains in fitness are redder.  
    The arc indicates the general direction relative to the start.
    
    Place the cursor over the symbol for more information.
    '''

    df = Statistics(s). \
        by_name(ActivityTopic, N.NAME). \
        by_name(ActivityCalculator, N.ACTIVE_DISTANCE, N.ACTIVE_TIME, N.TOTAL_CLIMB, N.DIRECTION,
                N.ASPECT_RATIO). \
        by_name(ActivityCalculator, N._delta(N.FITNESS_ANY), like=True).df
    df[N.DISTANCE_KM] = df[N.ACTIVE_DISTANCE]
    df['Duration'] = df[N.ACTIVE_TIME].map(format_seconds)
    df.loc[df[N.TOTAL_CLIMB].isna(), N.TOTAL_CLIMB] = 0

    calendar = Calendar(df, title='Distance, Fitness and Direction',
                        not_hover=[N.ACTIVE_DISTANCE, N.ACTIVE_TIME] +
                                  [column for column in df.columns if ':' in column])
    calendar.std_distance_fitness_direction()

    '''
    ## Fitness and Fatigue

    Better fitness has larger symbols.  When fatigue is higher symbols have "hotter" colours.
    
    Place the cursor over the symbol for more information.
    '''

    df = Statistics(s). \
        by_name(ActivityTopic, N.NAME). \
        by_name(ActivityCalculator, N.ACTIVE_DISTANCE, N.ACTIVE_TIME). \
        by_name(ActivityCalculator, N._delta(N.FITNESS_ANY), like=True).df
    df = Statistics(s). \
        by_name(ResponseCalculator, N.ANY_FATIGUE_ANY, N.ANY_FITNESS_ANY, like=True). \
        with_.drop_prefix(N.DEFAULT).into(df, tolerance='30m')
    fitness = sorted_numeric_labels(df.columns, N.FITNESS)[0]
    fatigue = sorted_numeric_labels(df.columns, N.FATIGUE)[0]
    df['FF Ratio'] = df[fatigue] / df[fitness]

    calendar = Calendar(df, title='Fitness and Fatigue', scale=18, border_month=0, border_day=0)
    calendar.set_size(fitness, min=0.1, gamma=0.5)
    calendar.set_palette('FF Ratio', magma(256), lo=0.5, hi=2, min=0)
    calendar.foreground('square', fill_alpha=1, line_alpha=0)
    calendar.show()

    '''
    ## Groups, Distance, Climb and Direction

    Larger distances have larger symbols.  Higher climbs are lighter.  
    The arc indicates the general direction relative to the start.
    Pastel backgrounds group similar rides.
    
    Place the cursor over the symbol for more information.
    '''

    dfa = Statistics(s). \
        by_name(ActivityTopic, N.NAME). \
        by_name(ActivityCalculator, N.ACTIVE_DISTANCE, N.ACTIVE_TIME, N.TOTAL_CLIMB, N.DIRECTION,
                N.ASPECT_RATIO). \
        by_name(ActivityCalculator, N._delta(N.FITNESS_ANY), like=True).df
    dfa[N.DISTANCE_KM] = df[N.ACTIVE_DISTANCE]
    dfa['Duration'] = df[N.ACTIVE_TIME].map(format_seconds)
    dfa.loc[dfa[N.TOTAL_CLIMB].isna(), N.TOTAL_CLIMB] = 0
    dfb = groups_by_time(s)
    if present(dfb, N.GROUP):
        dfb.loc[dfb[N.GROUP].isna(), N.GROUP] = -1
        df = dfa.join(dfb)

        calendar = Calendar(df, scale=15, border_day=0.1,
                            not_hover=[N.ACTIVE_DISTANCE, N.ACTIVE_TIME] +
                                      [column for column in df.columns if ':' in column])
        calendar.std_group_distance_climb_direction()



from logging import getLogger

from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql.functions import count

from .database import SystemConstant, Process, MappedDatabase, sqlite_uri, Database, Interval
from .support import SystemBase
from .tables.system import Progress, DirtyInterval
from ..commands.args import SYSTEM, DB_EXTN, DATA
from ..lib.utils import grouper

log = getLogger(__name__)


class System(MappedDatabase):

    def __init__(self, base):
        super().__init__(sqlite_uri(base, name=SYSTEM), SystemConstant, SystemBase)
        version = self.get_constant(SystemConstant.DB_VERSION, none=True)
        if version:
            log.info(f'Database version {version}')
        else:
            log.warning('Database unconfigured')

    def _sessionmaker(self):
        return sessionmaker(bind=self.engine, expire_on_commit=False)

    def get_constant(self, name, none=False):
        with self.session_context() as s:
            value = SystemConstant.from_name(s, name, none=none)
            return value

    def set_constant(self, name, value, force=False):
        log.debug(f'Setting {name}={value}')
        with self.session_context() as s:
            return SystemConstant.set(s, name, value, force=force)

    def delete_constant(self, name):
        log.debug(f'Deleting {name}')
        with self.session_context() as s:
            SystemConstant.delete(s, name)

    def get_process(self, owner, pid):
        with self.session_context() as s:
            return s.query(Process). \
                filter(Process.owner == owner,
                       Process.pid == pid).one()

    def run_process(self, owner, cmd, log_name):
        with self.session_context() as s:
            return Process.run(s, owner, cmd, log_name)  # todo change order

    def delete_process(self, owner, pid, delta_seconds=3):
        with self.session_context() as s:
            Process.delete(s, owner, pid, delta_seconds=delta_seconds)

    def delete_all_processes(self, owner, delta_seconds=3):
        with self.session_context() as s:
            Process.delete_all(s, owner, delta_seconds=delta_seconds)

    def exists_any_process(self, owner):
        with self.session_context() as s:
            return Process.exists_any(s, owner)

    def create_progress(self, name, delta_seconds=3):
        with self.session_context() as s:
            Progress.create(s, name, delta_seconds=delta_seconds)

    def update_progress(self, name, **kargs):
        with self.session_context() as s:
            Progress.update(s, name, **kargs)

    def remove_progress(self, name):
        with self.session_context() as s:
            Progress.remove(s, name)

    def get_percent(self, name):
        with self.session_context() as s:
            return Progress.get_percent(s, name)

    def wait_for_progress(self, name, timeout=60):
        with self.session_context() as s:
            return Progress.wait_for_progress(s, name, timeout=timeout)

    def get_database(self, uri=None):
        if not uri: uri = self.get_constant(SystemConstant.DB_URI, none=True)
        if uri:
            return Database(uri)
        else:
            log.warning('No database URI configured')
            return None

    def record_dirty_intervals(self, ids):
        if ids:
            with self.session_context() as s:
                # this doesn't have to be exact or thread safe
                known = set(s.query(DirtyInterval.interval_id).all())
                count = 0
                for id in ids:
                    if id not in known:
                        s.add(DirtyInterval(interval_id=id))
                        known.add(id)
                        count += 1
            if count: log.warning(f'Marked {count} intervals as dirty')

    def get_dirty_intervals(self):
        with self.session_context() as s:
            return s.query(DirtyInterval).all()

    def delete_dirty_intervals(self, intervals):
        dirty_ids = set(d.id for d in intervals)
        with self.session_context() as s:
            for ids in grouper(dirty_ids, 900):  # avoid limit in sqlite older versions
                s.query(DirtyInterval).filter(DirtyInterval.id.in_(ids)).delete(synchronize_session=False)


class Data:

    def __init__(self, base):
        self.__base = base
        self.__sys = None
        self.__db = None

    @property
    def base(self):
        return self.__base

    @property
    def sys(self):
        if not self.__sys:
            self.__sys = System(self.__base)
        return self.__sys

    @property
    def db(self):
        if not self.__db:
            self.__db = self.sys.get_database()
        return self.__db

    def reset(self):
        self.__sys = None
        self.__db = None




from logging import getLogger

from sqlalchemy import desc

from ..utils import ActivityJournalDelegate
from ....diary.model import optional_text, text
from ....sql import Achievement

log = getLogger(__name__)


class AchievementDelegate(ActivityJournalDelegate):

    @optional_text('Achievements')
    def read_journal_date(self, s, ajournal, date):
        achievements = list(self._achievements_for_journal(s, ajournal))
        if not achievements: achievements = [text('Get out and try harder!')]
        yield from achievements

    def _achievements_for_journal(self, s, ajournal):
        for achievement in s.query(Achievement). \
                filter(Achievement.activity_journal == ajournal). \
                order_by(desc(Achievement.sort)).limit(5).all():   # limit 5 for compact display
            yield text(achievement.text)

    def read_interval(self, s, interval):
        # todo?
        return
        yield





from logging import getLogger
from re import sub

from ..utils import ActivityJournalDelegate
from ...calculate.segment import SegmentCalculator
from ....names import Names
from ....diary.database import summary_column
from ....diary.model import value, text, optional_text
from ....sql.tables import SegmentJournal, Segment, StatisticJournal, StatisticName

log = getLogger(__name__)


def segments_for_activity(s, ajournal):
    return s.query(SegmentJournal). \
        filter(SegmentJournal.activity_journal == ajournal). \
        order_by(SegmentJournal.start).all()


class SegmentDelegate(ActivityJournalDelegate):

    @optional_text('Segments')
    def read_journal_date(self, s, ajournal, date):
        for sjournal in s.query(SegmentJournal).join(Segment). \
                filter(SegmentJournal.activity_journal_id == ajournal.id). \
                order_by(SegmentJournal.start).all():
            stats = [value(sub('^segment ', '', field.statistic_name.name), field.value,
                           units=field.statistic_name.units, measures=field.measures_as_model(date))
                     for field in (self.__field(s, date, sjournal, name)
                                   for name in (Names.SEGMENT_TIME, Names.SEGMENT_HEART_RATE))
                     if field]
            if stats:
                yield [text(sjournal.segment.title, tag='segment')] + stats

    @staticmethod
    def __field(s, date, sjournal, name):
        return StatisticJournal.at_date(s, date, name, SegmentCalculator, sjournal.activity_group,
                                        source_id=sjournal.id)
    @optional_text('Segments')
    def read_interval(self, s, interval):
        for segment in s.query(Segment).all():
            segment_rows = [list(summary_column(s, interval.schedule, interval.start, name))
                            for name in self.__names(s, Names.SEGMENT_TIME, Names.SEGMENT_HEART_RATE)]
            segment_rows = list(filter(bool, segment_rows))
            if segment_rows:
                yield [text(segment.title)] + segment_rows

    @staticmethod
    def __names(s, *names):
        for name in names:
            sname = s.query(StatisticName). \
                filter(StatisticName.name == name,
                       StatisticName.owner == SegmentCalculator).one_or_none()
            if sname:
                yield sname

    def read_interval(self, s, interval):
        # todo?
        return
        yield





from logging import getLogger

from .utils import MultiProcCalculator, SegmentJournalCalculatorMixin, DataFrameCalculatorMixin
from ..pipeline import OwnerInMixin
from ...data import present, linear_resample_time, Statistics
from ...names import N, Titles, Summaries as S, Units
from ...sql import SegmentJournal, StatisticJournalFloat

log = getLogger(__name__)

SJOURNAL = 'sjournal'


class SegmentCalculator(OwnerInMixin, SegmentJournalCalculatorMixin, DataFrameCalculatorMixin, MultiProcCalculator):

    def _delete(self, s):
        super()._delete(s)
        SegmentJournal.clean(s)

    def _read_dataframe(self, s, sjournal):
        from ..owners import SegmentReader
        return Statistics(s, activity_journal=sjournal.activity_journal). \
            by_name(SegmentReader, N.HEART_RATE).df

    def _calculate_stats(self, s, sjournal, df):
        stats = {SJOURNAL: sjournal,
                 N.SEGMENT_TIME: (sjournal.finish - sjournal.start).total_seconds()}
        if present(df, N.HEART_RATE):
            ldf = linear_resample_time(df, start=sjournal.start, finish=sjournal.finish)
            stats[N.SEGMENT_HEART_RATE] = ldf[N.HEART_RATE].mean()
        return stats

    def _copy_results(self, s, ajournal, loader, stats):
        sjournal = stats[SJOURNAL]
        loader.add(Titles.SEGMENT_TIME, Units.S, S.join(S.MIN, S.CNT, S.MSR), sjournal,
                   stats[N.SEGMENT_TIME], sjournal.start, StatisticJournalFloat)
        if N.SEGMENT_HEART_RATE in stats:
            loader.add(Titles.SEGMENT_HEART_RATE, Units.BPM, S.join(S.MAX, S.CNT, S.MSR),
                       sjournal, stats[N.SEGMENT_HEART_RATE], sjournal.start, StatisticJournalFloat)




from contextlib import contextmanager
from logging import getLogger
from re import sub
from sqlite3 import OperationalError, Connection

from sqlalchemy import create_engine, event, MetaData
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql.functions import count
from uritools import urisplit

from . import *
from .support import Base
from ..commands.args import NamespaceWithVariables, NO_OP, make_parser, DB_EXTN, base_system_path, DATA, ACTIVITY, BASE, \
    DB_VERSION, POSTGRESQL, SQLITE
from ..lib.io import data_hash
from ..lib.log import make_log_from_args

# mention these so they are "created" (todo - is this needed? missing tables seem to get created anyway)
Source,  Interval, Dummy, Composite, CompositeComponent
ActivityGroup, ActivityJournal, ActivityTimespan, ActivityBookmark
DiaryTopic, DiaryTopicJournal, DiaryTopicField,
ActivityTopic, ActivityTopicJournal, ActivityTopicField,
StatisticName, StatisticJournal, StatisticJournalInteger, StatisticJournalFloat, StatisticJournalText, StatisticMeasure
Segment, SegmentJournal
Pipeline
MonitorJournal
Constant, SystemConstant, Process
ActivitySimilarity, ActivityNearby
Timestamp


log = getLogger(__name__)

# https://stackoverflow.com/questions/13712381/how-to-turn-on-pragma-foreign-keys-on-in-sqlalchemy-migration-script-or-conf
@event.listens_for(Engine, "connect")
def fk_pragma_on_connect(dbapi_con, _con_record):
    if isinstance(dbapi_con, Connection):
        cursor = dbapi_con.cursor()

        def pragma(cmd):
            full_cmd = f'PRAGMA {cmd}'
            cursor.execute(full_cmd)

        pragma('foreign_keys=ON;')  # https://www.sqlite.org/pragma.html#pragma_foreign_keys
        pragma('temp_store=MEMORY;')  # https://www.sqlite.org/pragma.html#pragma_temp_store
        pragma('threads=4;')  # https://www.sqlite.org/pragma.html#pragma_threads
        pragma('cache_size=-1000000;')  # 1GB  https://www.sqlite.org/pragma.html#pragma_cache_size
        pragma('secure_delete=OFF;')  # https://www.sqlite.org/pragma.html#pragma_secure_delete
        pragma('journal_mode=WAL;')  # https://www.sqlite.org/wal.html
        pragma(f'busy_timeout={5 * 60 * 1000};')  # https://www.sqlite.org/pragma.html#pragma_busy_timeout
        cursor.close()


@event.listens_for(Engine, 'close')
def analyze_pragma_on_close(dbapi_con, _con_record):
    if isinstance(dbapi_con, Connection):
        cursor = dbapi_con.cursor()
        try:
            # this can fail if another process is using the database
            cursor.execute("PRAGMA optimize;")  # https://www.sqlite.org/pragma.html#pragma_optimize
        except OperationalError as e:
            log.debug("Optimize DB aborted (DB likely still in use)")
        finally:
            cursor.close()


def scheme(uri):
    return urisplit(uri).scheme


def sqlite_uri(base, read_only=False, name=ACTIVITY, version=DB_VERSION):
    path = base_system_path(base, subdir=DATA, file=name + DB_EXTN, version=version)
    uri = f'{SQLITE}:///{path}'
    if read_only: uri += '?mode=ro'
    return uri


def postgresql_uri(read_only=False, version=DB_VERSION):
    '''
    We no longer use base here.  It was a confused mess.  You can still have a database that depends on base
    because the system database still switches, so you can explicitly configure a different database uri.

    We use the default postgres schema because they cannot be managed within the URI.
    See also https://docs.sqlalchemy.org/en/13/dialects/postgresql.html#remote-schema-table-introspection-and-postgresql-search-path
    '''
    if read_only: log.warning('Read-only not supported yet for Postgres')
    name = f'{ACTIVITY}-{version}'
    return f'{POSTGRESQL}://postgres@localhost/{name}'


class DatabaseBase:

    def __init__(self, uri):
        self.uri = uri
        log.info('Using database at %s' % self.uri)
        options = {'echo': False}
        if POSTGRESQL == scheme(uri): options.update(executemany_mode="values")
        log.debug(f'Creating engine for {uri} with options {options}')
        self.engine = create_engine(uri, **options)
        self.session = self._sessionmaker()

    def _sessionmaker(self):
        return sessionmaker(bind=self.engine)

    def _no_schema(self, table):
        # https://stackoverflow.com/questions/33053241/sqlalchemy-if-table-does-not-exist
        return not self.engine.dialect.has_table(self.engine, table.__tablename__)

    @contextmanager
    def session_context(self):
        session = self.session()
        try:
            yield session
            session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def __str__(self):
        return f'{self.__class__.__name__} at {self.uri}'


class MappedDatabase(DatabaseBase):

    def __init__(self, uri, table, base):
        super().__init__(uri)
        if self._no_schema(table):
            log.info('Creating tables')
            base.metadata.create_all(self.engine)


class Database(MappedDatabase):

    # please create via sys.get_database !!

    def __init__(self, uri):
        super().__init__(uri, Source, Base)

    def no_data(self,):
        with self.session_context() as s:
            n_topics = s.query(count(DiaryTopic.id)).scalar()
            n_activities = s.query(count(ActivityGroup.id)).scalar()
            n_statistics = s.query(count(StatisticName.id)).scalar()
            return not (n_topics + n_activities + n_statistics)


def connect(args):
    '''
    Bootstrap from commandline-like args.
    '''
    from .system import System
    if len(args) == 1 and isinstance(args[0], str):
        args = args[0].split()
    elif args:
        args = list(args)
    else:
        args = []
    args.append(NO_OP)
    ns = NamespaceWithVariables(make_parser(with_noop=True).parse_args(args))
    make_log_from_args(ns)
    sys = System(ns[BASE])
    db = sys.get_database()
    return ns, db


class ReflectedDatabase(DatabaseBase):

    def __init__(self, *args, **kargs):
        super().__init__(*args, **kargs)
        self.meta = MetaData()
        self.meta.reflect(bind=self.engine)




from logging import getLogger

import numpy as np
import pandas as pd
from math import atan2, cos, sin, sqrt, pi

from .frame import linear_resample, median_dt, present, linear_resample_time
from ..lib.data import safe_dict
from ..names import Names

log = getLogger(__name__)

MAX_MINUTES = (5, 10, 30, 60, 90, 120, 180)


def round_km():
    yield from range(5, 21, 5)
    yield from range(25, 76, 25)
    yield from range(100, 251, 50)
    yield from range(300, 1001, 100)


@safe_dict
def active_stats(df):
    stats = {Names.ACTIVE_DISTANCE: 0, Names.ACTIVE_TIME: 0, Names.ACTIVE_SPEED: 0}
    for timespan in df[Names.TIMESPAN_ID].dropna().unique():
        slice = df.loc[df[Names.TIMESPAN_ID] == timespan]
        stats[Names.ACTIVE_DISTANCE] += slice[Names.DISTANCE].max() - slice[Names.DISTANCE].min()
        stats[Names.ACTIVE_TIME] += (slice.index.max() - slice.index.min()).total_seconds()
    stats[Names.ACTIVE_SPEED] = 3600 * stats[Names.ACTIVE_DISTANCE] / stats[Names.ACTIVE_TIME]
    return stats


@safe_dict
def copy_times(ajournal):
    return {Names.START: ajournal.start,
            Names.FINISH: ajournal.finish,
            Names.TIME: (ajournal.finish - ajournal.start).total_seconds()}


@safe_dict
def times_for_distance(df, km=None, delta=0.01):  # all units of km
    stats, km = {}, km or round_km()
    tmp = pd.DataFrame({Names.TIME: df.index}, index=df[Names.DISTANCE])
    tmp = tmp[~tmp.index.duplicated(keep='last')]
    t4d = pd.DataFrame({Names.TIME: (tmp[Names.TIME] - tmp[Names.TIME].iloc[0]).astype(np.int64) / 1e9},
                       index=df[Names.DISTANCE])
    lt4d = linear_resample(t4d, d=delta)
    for target in km:
        n = target / delta
        dlt4d = lt4d.diff(periods=n).dropna()
        if present(dlt4d, Names.TIME):
            stats[Names.MIN_KM_TIME % target] = dlt4d[Names.TIME].min()
            stats[Names.MED_KM_TIME % target] = dlt4d[Names.TIME].median()
    return stats


@safe_dict
def hrz_stats(df, zones=None):
    stats, zones = {}, zones or range(1, 8)
    if present(df, Names.HR_ZONE):
        ldf = linear_resample_time(df, with_timespan=True)
        hrz = pd.cut(ldf[Names.HR_ZONE], bins=zones, right=False).value_counts()
        dt, total = median_dt(ldf), hrz.sum()
        for interval, count in hrz.iteritems():
            zone = interval.left
            stats[Names.PERCENT_IN_Z % zone] = 100 * count / total
            stats[Names.TIME_IN_Z % zone] = dt * count
    return stats


@safe_dict
def max_mean_stats(df, params=((Names.POWER_ESTIMATE, Names.MAX_MEAN_PE_M),), mins=None, delta=10, zero=0):
    stats, mins = {}, mins or MAX_MINUTES
    ldf = linear_resample_time(df, dt=delta, with_timespan=True, keep_nan=True)
    for name, template in params:
        if name in ldf.columns:
            ldf.loc[ldf[Names.TIMESPAN_ID].isnull(), [name]] = zero
            cumsum = ldf[name].cumsum()
            for target in mins:
                n = (target * 60) // delta
                diff = cumsum.diff(periods=n).dropna()
                if present(diff, name):
                    stats[template % target] = diff.max() / n
        else:
            log.warning(f'Missing {name}')
    return stats


@safe_dict
def max_med_stats(df, params=((Names.HEART_RATE, Names.MAX_MED_HR_M),), mins=None, delta=10, gap=0.01):
    stats, mins = {}, mins or MAX_MINUTES
    ldf_all = linear_resample_time(df, dt=delta, with_timespan=False, add_time=False)
    ldf_all.interpolate('nearest')
    ldf_tstamp = ldf_all.loc[ldf_all[Names.TIMESPAN_ID].isin(df[Names.TIMESPAN_ID].unique())].copy()
    ldf_tstamp.loc[:, 'gap'] = ldf_tstamp.index.astype(np.int64) / 1e9
    ldf_tstamp.loc[:, 'gap'] = ldf_tstamp['gap'].diff()
    log.debug(f'Largest gap is {ldf_tstamp["gap"].max()}s')
    for target in mins:
        n = target * 60 // delta
        log.debug(f'Target {target}m is {n} samples (delta {delta}s)')
        splits, remain = [], ldf_all.copy()
        max_gap = max(gap * target * 60, 1.5 * delta)
        for after in ldf_tstamp.index[ldf_tstamp['gap'] > max_gap].tolist():
            before = ldf_tstamp.index[ldf_tstamp.index.get_loc(after)-1]
            splits.append(remain.loc[:before])
            remain = remain.loc[after:]
        splits.append(remain)
        log.debug(f'Split data into {len(splits)} sections for {target}m with max gap of {max_gap}s')
        for name, template in params:
            stat_name = template % target
            for split in splits:
                split['med'] = split[name].rolling(n).median()
                if present(split, 'med'):
                    max_med = split['med'].dropna().max()
                    if stat_name in stats:
                        stats[stat_name] = max(stats[stat_name], max_med)
                    else:
                        stats[stat_name] = max_med
    return stats


@safe_dict
def direction_stats(df):
    stats = {}
    if all(name in df.columns for name in (Names.SPHERICAL_MERCATOR_X, Names.SPHERICAL_MERCATOR_Y)):
        df = df.dropna(subset=[Names.SPHERICAL_MERCATOR_X, Names.SPHERICAL_MERCATOR_Y]).copy()
        if not df.empty:
            x0, y0 = df.iloc[0][Names.SPHERICAL_MERCATOR_X], df.iloc[0][Names.SPHERICAL_MERCATOR_Y]
            df.loc[:, 'dx'] = df[Names.SPHERICAL_MERCATOR_X] - x0
            df.loc[:, 'dy'] = df[Names.SPHERICAL_MERCATOR_Y] - y0
            # average position
            dx, dy = df['dx'].mean(), df['dy'].mean()
            x1, y1, d = x0 + dx, y0 + dy, sqrt(dx ** 2 + dy ** 2)
            theta = atan2(dy, dx)
            # change coords to centred on average position and perp / parallel to line to start
            df.loc[:, 'dx'] = df[Names.SPHERICAL_MERCATOR_X] - x1
            df.loc[:, 'dy'] = df[Names.SPHERICAL_MERCATOR_Y] - y1
            df.loc[:, 'u'] = df['dx'] * cos(theta) + df['dy'] * sin(theta)
            df.loc[:, 'v'] = df['dy'] * cos(theta) - df['dx'] * sin(theta)
            # convert from angle anti-clock from x axis to bearing
            stats[Names.DIRECTION] = 90 - 180 * theta / pi
            stats[Names.ASPECT_RATIO] = df['v'].std() / df['u'].std()
    return stats



from collections import defaultdict
from logging import getLogger
from os.path import sep, exists, join, isfile

from sqlalchemy_utils import database_exists

from .args import SOURCE, ACTIVITY, DB_EXTN, base_system_path, SEGMENTS, CONSTANTS, KIT, ACTIVITIES, DIARY, \
    infer_flags, ENGINE, SQLITE, POSTGRESQL, mm
from .read import DATA
from ..lib.log import Record
from ..lib.utils import clean_path
from ..import_.activity import import_activity
from ..import_.constant import import_constant
from ..import_.diary import import_diary
from ..import_.kit import import_kit
from ..import_.segment import import_segment
from ..sql.database import ReflectedDatabase, sqlite_uri, postgresql_uri, SystemConstant, scheme

log = getLogger(__name__)


def import_(args, data):
    '''
## import

    > ch2 import 0-30

Import data from a previous version (after starting a new version).
Data must be imported before any other changes are made to the database.

By default all types of data (diary, activities, kit, constants and segments) are imported.
Additional flags can enable or disable specific data types.

### Examples

    > ch2 import --diary 0-30

Import only diary entries.

    > ch2 import --disable --diary 0-30

Import everything but diary entries.
    '''
    flags = infer_flags(args, DIARY, ACTIVITIES, KIT, CONSTANTS, SEGMENTS)
    import_source(data, Record(log), args[SOURCE], args[ENGINE], flags=flags)


def infer_uri(base, source, engine, sys):
    if ':' in source:
        if engine:
            raise Exception(f'Do not specify engine with (what looks like) a URI')
        uri = source
    elif len(source) < 8:
        if not engine:
            current = sys.get_constant(SystemConstant.DB_URI, none=True)
            if not current:
                raise Exception(f'Specify {mm(SQLITE)} or {mm(POSTGRESQL)}')
            engine = scheme(current)
        if engine == SQLITE:
            uri = sqlite_uri(base, version=source)
        elif engine == POSTGRESQL:
            uri = postgresql_uri(version=source)
        else:
            raise Exception(f'Unknown engine {engine}')
    else:
        raise Exception(f'Unexpected version or URI {source}')
    if database_exists(uri):
        return uri
    else:
        raise Exception(f'No database at {uri}')


def import_source(data, record, source, engine=None, flags=None):
    # engine needed if source is not a URI
    with record.record_exceptions():
        uri = infer_uri(data.base, source, engine, data.sys)
        if flags is None: flags = defaultdict(lambda: True)
        old = ReflectedDatabase(uri)
        if not old.meta.tables:
            record.raise_(f'No tables found in {uri}')
        log.info(f'Importing data from {uri}')
        if flags[DIARY]: import_diary(record, old, data.db)
        if flags[ACTIVITIES]: import_activity(record, old, data.db)
        if flags[KIT]: import_kit(record, old, data.db)
        if flags[CONSTANTS]: import_constant(record, old, data.db)
        if flags[SEGMENTS]: import_segment(record, old, data.db)


def build_source_path(record, base, source):

    def nice_msg(template, source, path):
        msg = template
        if source != path: msg += f' ({path})'
        return msg

    database = ACTIVITY + DB_EXTN
    if sep not in source:
        path = base_system_path(base, subdir=DATA, file=database, version=source, create=False)
        if exists(path):
            log.info(nice_msg(f'{source} appears to be a version', source, path))
            return path
        else:
            log.warning(nice_msg(f'{source} is not a version', source, path))
    path = clean_path(source)
    if exists(path) and isfile(path):
        log.info(nice_msg(f'{source} exists', source, path))
        return path
    else:
        log.warning(f'{source} is not a database file ({path})')
    path = join(path, database)
    if exists(path) and isfile(path):
        log.info(nice_msg(f'{source} exists', source, path))
        return path
    else:
        log.warning(nice_msg(f'{source} is not a base directory', source, path))
    record.raise_(f'Could not find {source}')






from logging import getLogger

from ...pipeline.display.activity.utils import latest_activity, activity_times_by_group

log = getLogger(__name__)


class Analysis:

    @staticmethod
    def read_parameters(request, s):
        # odds and sods used to set menus in jupyter URLs
        latest = latest_activity(s)
        times_by_group = activity_times_by_group(s)
        all_activity_times = sorted(sum(times_by_group.values(), []), reverse=True)
        return {'all_activity_times': all_activity_times,
                'activity_times_by_group': times_by_group,
                'latest_activity_group': latest.activity_group.name if latest else None}



from logging import getLogger
from tempfile import TemporaryDirectory

import pandas as pd

from ch2.commands.read import read
from ch2.commands.args import bootstrap_dir, mm, DEV, V, m
from ch2.config.profile.default import default
from ch2.data import Names as N, Statistics
from ch2.pipeline.read import SegmentReader
from tests import LogTestCase

log = getLogger(__name__)


class TestPower(LogTestCase):

    def test_constant(self):

        with TemporaryDirectory() as f:

            bootstrap_dir(f, m(V), '5', mm(DEV), configurator=default)

            args, data = bootstrap_dir(f, m(V), '5', mm(DEV), 'read',
                                       'data/test/source/personal/2018-03-04-qdp.fit')
            read(args, data)

            with data.db.session_context() as s:
                stats = Statistics(s, activity_journal='2018-03-04 07:16:33', with_timespan=True). \
                    by_name(SegmentReader, N.LATITUDE, N.LONGITUDE, N.SPHERICAL_MERCATOR_X,
                            N.SPHERICAL_MERCATOR_Y, N.DISTANCE, N.ELEVATION, N.SPEED, N.CADENCE, N.HEART_RATE).df
                stats.describe()

                sepn = pd.Series(stats.index).diff().median()  # 7 secs
                start = stats.index.min()  # 2018-03-04 10:16:33+00:00
                finish = stats.index.max()  # 2018-03-04 16:34:51+00:00
                even = pd.DataFrame({'keep': True}, index=pd.date_range(start=start, end=finish, freq=sepn))
                both = stats.join(even, how='outer', sort=True)
                both.interpolate(method='index', limit_area='inside', inplace=True)
                both = both.loc[both['keep'] == True].drop(columns=['keep'])
                both = both.loc[both[N.TIMESPAN_ID].isin(stats[N.TIMESPAN_ID].unique())]
                both.describe()




from logging import getLogger

from .args import FORCE, LIKE, FINISH, START, WORKER, parse_pairs, KARG, UNLIKE, BASE
from ..sql.tables.pipeline import PipelineType
from ..pipeline.pipeline import run_pipeline

log = getLogger(__name__)


def calculate(args, data):
    '''
## calculate

    > ch2 calculate

Calculate any missing statistics.

    > ch2 calculate --force [START [FINISH]]

Delete statistics in the date range (or all, if omitted) and then calculate new values.

    > ch2 --dev calculate --like '%Activity%' --force 2020-01-01 -Kn_cpu=1

Calculate activity statistics from 2020 onwards in a single process for debugging.
    '''
    run_statistic_pipelines(data, force=args[FORCE], like=args[LIKE], unlike=args[UNLIKE],
                            start=args[START], finish=args[FINISH], worker=args[WORKER] is not None, id=args[WORKER],
                            **parse_pairs(args[KARG]))


def run_statistic_pipelines(data, force=False, like=tuple(), unlike=tuple(), start=None, finish=None,
                            worker=False, id=None, **kargs):
    run_pipeline(data, PipelineType.CALCULATE, force=force, like=like, unlike=unlike, start=start, finish=finish,
                 worker=worker, id=id, **kargs)




from .achievement import AchievementCalculator
from .activity import ActivityCalculator
from .elevation import ElevationCalculator
from .heart_rate import RestHRCalculator
from .impulse import ImpulseCalculator
from .kit import KitCalculator
from .monitor import MonitorCalculator
from .nearby import NearbyCalculator
from .response import ResponseCalculator
from .segment import SegmentCalculator
from .summary import SummaryCalculator



from collections import defaultdict
from glob import glob
from logging import getLogger
from os import makedirs
from os.path import basename, join, exists, dirname

from math import sqrt

from .calculate import run_statistic_pipelines
from .garmin import run_garmin
from ..commands.args import KIT, READ, BASE, FORCE, PATH, base_system_path, PERMANENT, WORKER, parse_pairs, \
    KARG, infer_flags, ACTIVITIES, CALCULATE
from ..lib.date import time_to_local_time, Y, YMDTHMS
from ..lib.io import data_hash, split_fit_path, touch
from ..lib.log import log_current_exception, Record
from ..lib.utils import clean_path
from ..lib.workers import ProgressTree, SystemProgressTree
from ..pipeline.pipeline import run_pipeline
from ..pipeline.read.activity import ActivityReader
from ..pipeline.read.monitor import MonitorReader
from ..pipeline.read.utils import AbortImportButMarkScanned
from ..sql import KitItem, FileHash, ActivityJournal, PipelineType

log = getLogger(__name__)


TYPE = 'type'
TIME = 'time'
STREAM = 'stream'
DATA = 'data'
EXTRA = 'extra'
HASH = 'hash'
NAME = 'name'
ACTIVITY = 'activity'
MONITOR = 'monitor'
SPORT = 'sport'
DIR = 'dir'
DOT_FIT = '.fit'
READ_PATH = 'read-path'
WRITE_PATH = 'write-path'

FLAGS = (ACTIVITIES, MONITOR, CALCULATE)


def read(args, data):
    '''
## read

    > ch2 read --kit ITEM [ITEM...] -- PATH [PATH ...]

Read FIT files, storing the data in the permanent store on the file system,
then scan their contents, adding entries to the database, and finally calculating associated statistics.
Both monitor and activity files are accepted.

Scanning and calculation of activities can be disabled with --disable, and individual steps can
be enabled / disabled by name.

### Examples

    > ch2 read --kit cotic -- ~/fit/2018-01-01.fit

will store the given file, add activity data to the database (associated with the kit 'cotic'), check for
new monitor data, and update statistics.

    > ch2 read --calculate

is equivalent to `ch2 calculate` (so will not store and files, will not read data, but wil calculate statistics).

    > ch2 read --disable --calculate [PATH ...]

will read files and add their contents to the database, but not calculate statistics.

    > ch2 read --disable [PATH ...]

will read files (ie copy them to the permanent store), but do no other processing.

Note: When using bash use `shopt -s globstar` to enable ** globbing.
    '''
    if args[WORKER]:
        run_pipeline(data, None, paths=args[PATH], force=args[FORCE], worker=bool(args[WORKER]),
                     id=args[WORKER], **parse_pairs(args[KARG]))
    else:
        record = Record(log)
        flags = infer_flags(args, *FLAGS)
        nfiles, files = open_files(args[PATH])
        upload_files_and_update(record, data, files=files, nfiles=nfiles, force=args[FORCE],
                                items=args[KIT], flags=flags, **parse_pairs(args[KARG]))


class SkipFile(Exception):
    pass


def open_files(paths):
    # converts a list of paths to a map of file names to file dicts with open streams
    n = len(paths)

    def files():
        # use an iterator here to avoid opening too many files at once
        for path in paths:
            path = clean_path(path)
            if exists(path):
                name = basename(path)
                log.debug(f'Reading {path}')
                stream = open(path, 'rb')
                yield {STREAM: stream, NAME: name, READ_PATH: path}
            else:
                log.warning(f'File no longer exists at {path}')

    return n, files()


def check_items(s, items):
    for item in items:
        if not s.query(KitItem).filter(KitItem.name == item).one_or_none():
            raise Exception(f'Kit item {item} does not exist')


def read_file(file):
    # add DATA to dicts with STREAM
    log.debug(f'Reading {file[NAME]}')
    file[DATA] = file[STREAM].read()
    file[STREAM].close()


def hash_file(file):
    # add HASH to dicts with DATA
    log.debug(f'Hashing {file[NAME]}')
    file[HASH] = data_hash(file[DATA])
    log.debug(f'Hash of {file[NAME]} is {file[HASH]}')


def check_path(file):
    path, name = file[WRITE_PATH], file[NAME]
    gpath, _ = split_fit_path(path)
    match = glob(gpath)
    if match:
        touch(match[0])  # trigger re-processing
        raise SkipFile(f'A file already exists for {name} at {match[0]}')
    log.debug(f'File {name} cleared for path {path}')


def check_hash(s, file):
    hash, path, name = file[HASH], file[WRITE_PATH], file[NAME]
    file_hash = s.query(FileHash).filter(FileHash.hash == hash).one_or_none()
    if file_hash and file_hash.file_scan:
        raise SkipFile(f'A file was already processed with the same hash as {name} '
                       f'({hash} at {file_hash.file_scan.path})')
    log.debug(f'File {name} cleared for hash {hash}')


def check_file(s, file):
    check_path(file)
    check_hash(s, file)


def parse_fit_data(file, items=None):
    try:
        # add TIME and TYPE and EXTRA (and maybe SPORT) given (fit) DATA and NAME
        try:
            records = MonitorReader.parse_records(file[DATA])
            file[TIME] = MonitorReader.read_first_timestamp(file[NAME], records)
            file[TYPE] = MONITOR
            # don't use '-' here or it will be treated as kit in path matching
            file[EXTRA] = ':' + file[HASH][0:5]
            log.debug(f'File {file[NAME]} contains monitor data')
        except AbortImportButMarkScanned:
            records = ActivityReader.parse_records(file[DATA])
            file[TIME] = ActivityReader.read_first_timestamp(file[NAME], records)
            file[SPORT] = ActivityReader.read_sport(file[NAME], records)
            file[TYPE] = ACTIVITY
            file[EXTRA] = '-' + ','.join(items) if items else ''
            log.debug(f'File {file[NAME]} contains activity data')
    except Exception as e:
        log_current_exception()
        raise Exception(f'Could not parse {file[NAME]} as a fit file')


def build_path(data_dir, file):
    # add PATH given TIME, TYPE, SPORT and EXTRA
    date = time_to_local_time(file[TIME], fmt=YMDTHMS)
    year = time_to_local_time(file[TIME], fmt=Y)
    file[DIR] = join(data_dir, file[TYPE], year)
    if SPORT in file: file[DIR] = join(file[DIR], file[SPORT])
    file[WRITE_PATH] = join(file[DIR], date + file[EXTRA] + DOT_FIT)
    log.debug(f'Target directory is {file[DIR]}')


def write_file(file):
    try:
        dir = dirname(file[WRITE_PATH])
        if not exists(dir):
            log.debug(f'Creating {dir}')
            makedirs(dir)
        if exists(file[WRITE_PATH]):
            log.warning(f'Overwriting data at {file[WRITE_PATH]}')
        with open(file[WRITE_PATH], 'wb') as out:
            log.info(f'Writing {file[NAME]} to {file[WRITE_PATH]}')
            out.write(file[DATA])
    except Exception as e:
        log_current_exception()
        raise Exception(f'Could not save {file[NAME]}')


def upload_files(record, data, files=tuple(), nfiles=1, items=tuple(), progress=None):
    try:
        local_progress = ProgressTree(len(files), parent=progress)
    except TypeError:
        local_progress = ProgressTree(nfiles, parent=progress)
    with record.record_exceptions():
        with data.db.session_context() as s:
            data_dir = base_system_path(data.base, version=PERMANENT)
            check_items(s, items)
            for file in files:
                with local_progress.increment_or_complete():
                    try:
                        read_file(file)
                        hash_file(file)
                        parse_fit_data(file, items=items)
                        build_path(data_dir, file)
                        check_file(s, file)
                        write_file(file)
                        record.info(f'Uploaded {file[NAME]} to {file[WRITE_PATH]}')
                    except SkipFile as e:
                        record.warning(e)
            local_progress.complete()  # catch no files case


def upload_files_and_update(record, data, files=tuple(), nfiles=1, force=False, items=tuple(),
                            flags=None, **kargs):
    # this expects files to be a list of maps from name to stream (or an iterator, if nfiles provided)
    if not flags:
        flags = defaultdict(lambda: True)
    n_options = sum(1 if flags[name] else 0 for name in FLAGS)
    if flags[MONITOR]: n_options += 2
    progress = SystemProgressTree(data.sys, READ, [1] * (n_options + 1))
    log.info(f'Uploading files')
    try:
        upload_files(record, data, files=files, nfiles=nfiles, items=items, progress=progress)
        # todo - add record to pipelines?
        if flags[ACTIVITIES]:
            log.info('Running activity pipelines')
            run_pipeline(data, PipelineType.READ_ACTIVITY, force=force, progress=progress, **kargs)
        if flags[MONITOR]:
            # run before and after so we know what exists before we update, and import what we read
            log.info('Running monitor pipelines')
            run_pipeline(data, PipelineType.READ_MONITOR, force=force, progress=progress, **kargs)
            with data.db.session_context() as s:
                try:
                    log.info('Running Garmin download')
                    run_garmin(data.sys, s, base=data.base, progress=progress)
                except Exception as e:
                    log.warning(f'Could not get data from Garmin: {e}')
            log.info('Running monitor pipelines (again)')
            run_pipeline(data, PipelineType.READ_MONITOR, force=force, progress=progress, **kargs)
        if flags[CALCULATE]:
            log.info('Running statistics pipelines')
            run_statistic_pipelines(data, force=force, progress=progress, **kargs)
    finally:
        progress.complete()




from ch2.names import Names, titles_for_names
from tests import LogTestCase


class TestNames(LogTestCase):

    def test_names(self):
        self.assertEqual(Names.HEART_RATE, 'heart-rate')
        self.assertEqual(Names.DELTA_AIR_SPEED_2, 'd-air-speed2')
        self.assertEqual(Names._avg('foo'), 'avg-foo')

    def test_titles_for_names(self):
        self.assertEqual(list(titles_for_names('A%C', ['abc', 'ABC', 'aBc'])), ['AbC', 'ABC'])




import datetime as dt
from collections import defaultdict
from logging import getLogger
from re import escape

from sqlalchemy import union, intersect, not_
from sqlalchemy.orm import aliased

from ..lib import local_time_to_time, to_time
from ..lib.peg import transform, choice, pattern, sequence, Recursive, drop, exhaustive, single
from ..lib.utils import timing
from ..names import UNDEF
from ..sql import ActivityJournal, StatisticName, StatisticJournalType, ActivityGroup, ActivityTopicJournal, Source, \
    FileHash, StatisticJournal
from ..sql.tables.statistic import STATISTIC_JOURNAL_CLASSES
from ..sql.types import lookup_cls

log = getLogger(__name__)


def pat(regexp, trans=None):
    parser = pattern(r'\s*' + regexp + r'\s*')
    if trans:
        parser = transform(parser, trans)
    return parser


def lit(text, trans=None):
    return pat('(' + escape(text) + ')', trans=trans)


def unexp(values):
    try:
        val, exp = values
        return [float(val) * 10 ** int(exp)]
    except:
        return [float(values[0])]


def to_tuple(l):
    return [tuple(l)]


def join(a, sep, b):
    return transform(sequence(a, sep, b), to_tuple)


def flip(l):
    value, op, name = l[0]
    return [(name, {'=': '=', '!=': '!=', '<': '>', '>': '<', '>=': '<=', '<=': '>='}[op], value)]


AND, OR, NULL = 'and', 'or', 'null'
# no wildcards in group (no use case)
name = pat(r'((?:(?:[A-Z%][A-Za-z0-9\-%]*)?\.)?[a-z%][a-z0-9\-%]*(?::[a-z][a-z0-9\-]*)?)')
# important that the different value types are exclusive to avoid ambiguities
number = choice(pat(r'(-?\d+)', lambda x: [int(i) for i in x]),
                pat(r'(-?\d*\.\d+)', lambda x: [float(i) for i in x]),
                pat(r'(-?\d+\.\d*)', lambda x: [float(i) for i in x]),
                pat(r'([-+]?\d*\.?\d+)[eE]([-+]?\d+)', unexp))
string = choice(pat(r'"((?:[^"\\]|\\.)*)"'), pat(r"'((?:[^'\\]|\\.)*)'"))
date = r'\d{4}-\d{2}-\d{2}'
time = r'\d{2}(?::\d{2}(?::\d{2})?)?'
datetime = pat('(' + date + '(?: ' + time + ')?)', lambda x: [local_time_to_time(t) for t in x])
utc = pat('(' + date + 'T' + time + ')', lambda x: [to_time(t) for t in x])
null = transform(lit(NULL), lambda x: [None])
value = choice(number, string, datetime, utc, null)

operator = choice(*[lit(cmp) for cmp in ['=', '!=', '<', '>', '<=', '>=']])
comparison = choice(join(name, operator, value), transform(join(value, operator, name), flip))

term = Recursive()

and_comparison = Recursive()
and_comparison.calls(choice(term, join(term, lit(AND), and_comparison)))

or_comparison = Recursive()
or_comparison.calls(choice(and_comparison, join(and_comparison, lit(OR), or_comparison)))

parens = transform(sequence(drop(lit('(')), or_comparison, drop(lit(')'))))
term.calls(choice(comparison, parens))

constraint = single(exhaustive(choice(or_comparison, parens)))


def trace(f):
    # help trace source of problems in sql generation
    def wrapper(*args, **kargs):
        result = f(*args, **kargs)
        try:
            sql = result[0]
        except:
            sql = result
        # log.debug(f'{f.__name__}: {sql}')
        return result
    return wrapper


def constrained_sources(s, query, conversion=None):
    with timing('parse AST'):
        ast = constraint(query)[0]
    log.debug(f'AST: {ast}')
    attrs = set()
    with timing('check AST'):
        check_constraints(s, ast, attrs)
    log.debug('Checked constraints')
    with timing('build SQL'):
        q = build_source_query(s, ast, attrs, conversion=conversion)
    log.debug(f'Query: {q}')
    with timing('execute SQL'):
        return q.all()


def check_constraints(s, ast, attrs):
    l, op, r = ast
    if op in (AND, OR):
        check_constraints(s, l, attrs)
        check_constraints(s, r, attrs)
    else:
        check_constraint(s, ast, attrs)


def infer_types(value):
    if value is None: return list(StatisticJournalType)
    if isinstance(value, str): return [StatisticJournalType.TEXT]
    if isinstance(value, dt.datetime): return [StatisticJournalType.TIMESTAMP]
    return [StatisticJournalType.FLOAT, StatisticJournalType.INTEGER]


def check_constraint(s, ast, attrs):
    qname, op, value = ast
    try:
        check_statistic_name(s, qname, value)
    except Exception as e1:
        try:
            check_source_property(qname, value)
            attrs.add(qname)
        except Exception as e2:
            log.error(e1)
            log.error(e2)
            raise Exception(f'{qname} {op} {value} could not be validated')


def check_statistic_name(s, qname, value):
    owner, name, _ = StatisticName.parse(qname)
    q = s.query(StatisticName.id). \
        filter(StatisticName.name.like(name),
               StatisticName.statistic_journal_type.in_(infer_types(value)))
    if owner: q = q.filter(StatisticName.owner.like(owner))
    if not q.count():
        raise Exception(f'No match for statistic {qname} for type {value.__class__.__name__}')


def check_source_property(qname, value):
    cls, attr = parse_property(qname)
    # todo - check value type somehow


@trace
def build_source_query(s, ast, attrs, conversion=None):
    constraints = build_constraints(s, ast, attrs, conversion=conversion).cte()
    return s.query(Source).filter(Source.id.in_(constraints))


@trace
def build_constraints(s, ast, attrs, conversion=None):
    l, op, r = ast
    if op in (AND, OR):
        lcte = build_constraints(s, l, attrs, conversion=conversion)
        rcte = build_constraints(s, r, attrs, conversion=conversion)
        return build_join(op, lcte, rcte)
    else:
        constraint, null = build_constraint(s, ast, attrs, bool(conversion))
        if conversion: constraint = conversion(s, constraint, null)
        return constraint


@trace
def build_join(op, lcte, rcte):
    if op == OR:
        return aliased(union(lcte, rcte)).select()
    else:
        return aliased(intersect(lcte, rcte)).select()


@trace
def build_constraint(s, ast, attrs, with_conversion):
    qname, op, value = ast
    if qname in attrs:
        return build_property(s, ast), False
    else:
        return build_comparisons(s, ast, with_conversion)


def build_property(s, ast):
    qname, op, value = ast
    cls, attr = parse_property(qname)
    op_attr = get_op_attr(op, value)
    q = s.query(cls.id)
    if attr == 'nlike':
        q = q.filter(not_(getattr(cls, attr).ilike(value)))
    else:
        q = q.filter(getattr(getattr(cls, attr), op_attr)(value))
    return q


@trace
def build_comparisons(s, ast, with_conversion):
    qname, op, value = ast
    owner, name, group = StatisticName.parse(qname, default_activity_group=UNDEF)
    if value is None:
        if op == '=':
            return get_source_ids_for_null(s, owner, name, group, with_conversion), True
        else:
            return aliased(union(*[get_source_ids(s, owner, name, op, value, group, type)
                                   for type in StatisticJournalType
                                   if type != StatisticJournalType.STATISTIC])).select(), False
    elif isinstance(value, str):
        return get_source_ids(s, owner, name, op, value, group, StatisticJournalType.TEXT), False
    elif isinstance(value, dt.datetime):
        return get_source_ids(s, owner, name, op, value, group, StatisticJournalType.TIMESTAMP), False
    else:
        qint = get_source_ids(s, owner, name, op, value, group, StatisticJournalType.INTEGER)
        qfloat = get_source_ids(s, owner, name, op, value, group, StatisticJournalType.FLOAT)
        return aliased(union(qint, qfloat)).select(), False


def get_op_attr(op, value):
    attrs = {'=': '__eq__', '!=': '__ne__', '>': '__gt__', '>=': '__ge__', '<': '__lt__', '<=': '__le__'}
    if isinstance(value, str): attrs.update({'=': 'ilike', '!=': 'nlike'})
    return attrs[op]


def get_source_ids(s, owner, name, op, value, group, type):
    op_attr = get_op_attr(op, value)
    statistic_journal = STATISTIC_JOURNAL_CLASSES[type]
    q = s.query(Source.id). \
        join(statistic_journal). \
        join(StatisticName). \
        filter(StatisticName.name.like(name))
    if owner:
        q = q.filter(StatisticName.owner.like(owner))
    if group is not UNDEF:
        if group:
            q = q.join(ActivityGroup).filter(ActivityGroup.name.ilike(group))
        else:
            q = q.filter(Source.activity_group_id == None)
    if op_attr == 'nlike':  # no way to negate like in a single attribute
        q = q.filter(not_(statistic_journal.value.ilike(value)))
    else:
        q = q.filter(getattr(statistic_journal.value, op_attr)(value))
    return q


def get_source_ids_for_null(s, owner, name, group, with_conversion):
    q = s.query(StatisticJournal.source_id). \
        join(StatisticName). \
        filter(StatisticName.name.like(name))
    if owner:
        q = q.filter(StatisticName.owner.like(owner))
    if group is not UNDEF:
        if group:
            q = q.join(ActivityGroup).filter(ActivityGroup.name.ilike(group))
        else:
            q = q.join(Source).filter(Source.activity_group_id == None)
    if with_conversion:
        # will invert later (in conversion)
        return q
    else:
        return s.query(Source.id).filter(not_(Source.id.in_(q)))


def parse_property(qname):
    if ':' in qname:
        raise Exception(f'{qname} is not a valid property (contains :)')
    if '.' in qname:
        cls, attr = qname.rsplit('.', 1)
        if not cls: cls = 'Source'
    else:
        cls = 'Source'
        attr = qname
    try:
        clz = lookup_cls(cls)
    except:
        cls = 'ch2.sql.' + cls
        clz = lookup_cls(cls)
    getattr(clz, attr)
    log.info(f'Parsed {qname} as {clz}.{attr}')
    return clz, attr


class Intersect(object):
    pass


@trace
def activity_conversion(s, source_ids, null):
    # convert the query that gives any source id and select either those that are activities directly,
    # or activities associated with a topic (eg user entered activity notes)

    # for most queries, we have some source IDs and we want to know if they are activityjournal ids
    # (which we pass through) or activitytopicjournal ids (in which case we convert to activityjournal).
    source_ids = source_ids.cte()
    q_direct = s.query(ActivityJournal.id). \
        filter(ActivityJournal.id.in_(source_ids))
    q_via_topic = s.query(ActivityJournal.id). \
        join(FileHash). \
        join(ActivityTopicJournal). \
        filter(ActivityTopicJournal.id.in_(source_ids))
    q = aliased(union(q_direct, q_via_topic)).select()

    if null:
        # for 'is null' queries we are really asking if the data are missing (since values are not null constrained)
        # so we find what does exist and then invert it.  that inversion has to happen avter conversion
        return s.query(ActivityJournal.id).filter(not_(ActivityJournal.id.in_(q)))
    else:
        return q


def group_by_type(sources):
    groups = defaultdict(list)
    for source in sources:
        groups[type(source)].append(source)
    return groups


def sort_sources(sources):
    def key(source):
        for attr in 'start', 'time', 'id':
            if hasattr(source, attr):
                return getattr(source, attr)
    return sorted(sources, key=key)


def sort_groups(groups):
    return {key: sort_sources(values) for key, values in groups.items()}




from logging import getLogger

from ch2.pipeline.loader import SqliteLoader

log = getLogger(__name__)


def unlock(args, data):
    '''
## unlock

    > ch2 unlock

Remove the "dummy" entry from the SQLite database that is used to coordinate locking across processes.

This should not be needed in normal use.  DO NOT use when worker processes are still running.
Has no effect when used with PostgreSQL.
    '''
    # todo - sqlite only
    with data.db.session_context() as s:
        log.info('Removing dummy entry...')
        SqliteLoader.unlock(s)
        log.info('Removed (if present)')



import datetime as dt
from collections import namedtuple
from json import loads
from logging import getLogger

import numpy as np
from math import log10
from sqlalchemy import distinct
from sqlalchemy.sql.functions import count

from .utils import UniProcCalculator
from ..pipeline import LoaderMixin, OwnerInMixin
from ..read.segment import SegmentReader
from ...data import Statistics, present
from ...data.response import sum_to_hour, calc_response
from ...lib.date import round_hour, to_time, local_date_to_time, now
from ...names import Names as N, SPACE
from ...sql import StatisticJournal, Composite, StatisticName, Source, Constant, CompositeComponent, \
    StatisticJournalFloat
from ...sql.tables.source import SourceType
from ...sql.utils import add

log = getLogger(__name__)

Response = namedtuple('Response', 'src_owner, title, tau_days, start, scale')

SCALED = 'scaled'


class ResponseCalculator(OwnerInMixin, LoaderMixin, UniProcCalculator):
    '''
    this is hard to do correctly, incrementally.

    for now, we do all or nothing.

    first, we check if the current solution is complete:
    * all sources used (no need to check for gaps - composite chaining should do that)
    * within 3 hours of the current time (if not, we regenerate, padding with zeroes)

    if not complete, we regenerate the whole damn thing.
    '''

    def __init__(self, *args, response_constants=None, prefix=None, **kargs):
        self.response_constant_names = self._assert('response_constants', response_constants)
        self.prefix = self._assert('prefix', prefix)
        super().__init__(*args, **kargs)

    def _startup(self, s):
        super()._startup(s)
        self.response_constants = [Constant.from_name(s, name) for name in self.response_constant_names]
        self.responses = [Response(**loads(constant.at(s).value)) for constant in self.response_constants]

    def _delete(self, s):
        start, finish = self._start_finish(local_date_to_time)
        if start or self.force:
            if finish:
                log.warning(f'Ignoring finish - will delete all impulses from {start} onwards')
            self._delete_from(s, start)

    def _delete_from(self, s, start=None):
        composite_ids = s.query(Composite.id). \
            join(StatisticJournal, Composite.id == StatisticJournal.source_id). \
            join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
            filter(StatisticName.owner == self.owner_out)
        if start:
            composite_ids = composite_ids.filter(StatisticJournal.time >= start)
        log.debug(f'Delete query: {composite_ids}')
        n = s.query(count(Source.id)). \
            filter(Source.id.in_(composite_ids)). \
            scalar()
        if n:
            log.warning(f'Deleting {n} Composite sources ({start} onwards)')
            s.query(Source). \
                filter(Source.id.in_(composite_ids)). \
                delete(synchronize_session=False)
            s.commit()
            Composite.clean(s)

    def _missing(self, s):
        # clean out any gaps by unzipping the chained composites
        Composite.clean(s)
        # range we expect data for
        finish = round_hour(dt.datetime.now(tz=dt.timezone.utc), up=True)
        missing_sources = self.__missing_sources(s)
        missing_recent = any(self.__missing_recent(s, constant.short_name, finish)
                             for constant in self.response_constants)
        if missing_recent or missing_sources:
            if missing_recent: log.info('Incomplete coverage (so will re-calculate)')
            if missing_sources: log.info('Additional sources (so will re-calculate)')
            self._delete_from(s)
            start = round_hour(self.__start(s), up=False)
            return [(start, finish)]
        else:
            return []

    def __missing_recent(self, s, constant, now):
        log.debug('Searching for missing recent')
        finish = s.query(StatisticJournal.time). \
            join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
            filter(StatisticName.name == self.prefix + SPACE + constant,
                   StatisticName.owner == self.owner_out). \
            order_by(StatisticJournal.time.desc()).limit(1).one_or_none()
        if finish is None:
            log.debug(f'Missing data for {constant}')
            return True
        if (now - finish[0]).total_seconds() > 3 * 60 * 60:
            log.debug(f'Insufficient data for {constant}')
            return True
        return False

    def __missing_sources(self, s):
        log.debug('Searching for missing sources')
        available = s.query(count(distinct(Source.id))). \
            join(StatisticJournal). \
            join(StatisticName). \
            filter(StatisticName.name == self.prefix + SPACE + N.HR_IMPULSE_10)
        used = s.query(count(distinct(Source.id))). \
            join(CompositeComponent, CompositeComponent.input_source_id == Source.id). \
            join(Composite, Composite.id == CompositeComponent.output_source_id). \
            join(StatisticJournal, StatisticJournal.source_id == Composite.id). \
            join(StatisticName). \
            filter(StatisticName.owner == self.owner_out,
                   Source.type == SourceType.ACTIVITY)
        n_avaialble = available.scalar()
        n_used = used.scalar()
        log.debug(f'Using {n_used} of {n_avaialble} sources')
        return n_used != n_avaialble

    def __start(self, s):
        # avoid constants defined at time 0
        return s.query(StatisticJournal.time). \
            filter(StatisticJournal.time > 1.0). \
            order_by(StatisticJournal.time.asc()). \
            limit(1).one()[0]  # scalar can return None

    def _run_one(self, s, missed):
        data = self.__read_data(s)
        if N.HR_IMPULSE_10 in data.columns and N.COVERAGE in data.columns:
            # coverage is calculated by the loader and seems to reflect records that have location data but not
            # HR data.  so i guess it makes sense to scale.
            # i don't remember why / when i added this - it might have been when using an optical monitor?
            # it seems like a relatively small effect in most cases.
            data.loc[now()] = {N.HR_IMPULSE_10: 0.0, N._src(N.HR_IMPULSE_10): None, N.COVERAGE: 100}
            data[SCALED] = data[N.HR_IMPULSE_10] * 100 / data[N.COVERAGE]
            all_sources = list(self.__make_sources(s, data))
            for constant, response in zip(self.response_constants, self.responses):
                name = self.prefix + SPACE + constant.short_name
                log.info(f'Creating values for {response.title} ({name})')
                hr3600 = sum_to_hour(data, SCALED)
                params = (log10(response.tau_days * 24),
                          log10(response.start) if response.start > 0 else 1)
                result = calc_response(hr3600, params) * response.scale
                loader = self._get_loader(s, add_serial=False)
                source, sources = None, list(all_sources)
                for time, value in result.iteritems():
                    # the sources are much more spread out than the response, which is calculated every hour so
                    # that it is smooth.  so we only increment the source when necessary.
                    skipped = 0
                    while sources and time >= sources[0][0]:
                        source = sources.pop(0)[1]
                        skipped += 1
                        if skipped > 1:
                            log.warning(f'Skipping multiple sources at {time}')
                    loader.add(name, None, None, source, value, time,
                               StatisticJournalFloat, title=response.title,
                               description=f'The SHRIMP response for a decay of {response.tau_days} days')
                loader.load()

    def __read_data(self, s):
        from ..owners import ImpulseCalculator
        name = self.prefix + SPACE + N.HR_IMPULSE_10
        df = Statistics(s, with_source=True).by_name(ImpulseCalculator, name).with_. \
            rename({name: N.HR_IMPULSE_10, N._src(name): N._src(N.HR_IMPULSE_10)}).df
        name = N._cov(N.HEART_RATE)
        df = Statistics(s).by_name(SegmentReader, name).with_. \
            rename({name: N.COVERAGE}).into(df, tolerance='10s')
        if present(df, N.COVERAGE):
            df[N.COVERAGE].fillna(axis='index', method='ffill', inplace=True)
            df[N.COVERAGE].fillna(100, axis='index', inplace=True)
        return df

    def __make_sources(self, s, data):
        # this chains forwards from zero, adding a new composite for each new impulse source.
        # in theory we could reconstruct only missing entries.
        log.info('Creating sources')
        name = N._src(N.HR_IMPULSE_10)
        prev = add(s, Composite(n_components=0))
        yield to_time(0.0), prev
        # find times where the source changes
        changes = data.loc[data[name].ne(data[name].shift())]
        for time, row in changes.iterrows():
            id = row[name]
            if not np.isnan(id):
                composite = add(s, Composite(n_components=2))
                add(s, CompositeComponent(input_source_id=id, output_source=composite))
            else:
                composite = add(s, Composite(n_components=1))
            add(s, CompositeComponent(input_source=prev, output_source=composite))
            yield time, composite
            prev = composite
        s.commit()




import datetime as dt
from collections import defaultdict
from logging import getLogger

import numpy as np
from sqlalchemy import desc, and_, or_, func
from sqlalchemy.orm import aliased
from sqlalchemy.sql.functions import count

from .utils import AbortImport, AbortImportButMarkScanned, MultiProcFitReader
from ..loader import SqliteLoader, PostgresqlLoader
from ..pipeline import LoaderMixin
from ...commands.args import mm, FORCE, READ, SQLITE, POSTGRESQL
from ...data.frame import read_query
from ...fit.format.records import fix_degrees, unpack_single_bytes, merge_duplicates
from ...fit.profile.profile import read_fit
from ...lib import log_current_exception
from ...lib.date import time_to_local_date, format_time
from ...names import N, T, Units
from ...sql.database import StatisticJournalType, Source
from ...sql.tables.monitor import MonitorJournal
from ...sql.tables.statistic import StatisticJournalInteger, StatisticName, StatisticJournal
from ...sql.utils import add

log = getLogger(__name__)

ACTIVITY_TYPE_ATTR = 'activity_type'
HEART_RATE_ATTR = 'heart_rate'
MONITORING_ATTR = 'monitoring'
MONITORING_INFO_ATTR = 'monitoring_info'
STEPS_ATTR = 'steps'


def missing_dates(s, force=False):
    # we don't try to be perfect here.  the idea is that it's called to get the latest
    # updates, rather than fill in all gaps (do the bulk download thing for that).
    # we also don't try to get fractional data.
    # and as for timezones... we just assume garmin uses the local timezone.
    latest = s.query(MonitorJournal).order_by(desc(MonitorJournal.start)).limit(1).one_or_none()
    if latest is None:
        log.warning('No existing monitor data - ' +
                    'do a bulk download instead: https://www.garmin.com/en-US/account/datamanagement/')
        return
    # find the mid-point to avoid any problems with timezones and edge cases
    # (these files don't span more than a day)
    seconds = (latest.finish - latest.start).total_seconds() / 2
    start = time_to_local_date(latest.start + dt.timedelta(seconds=seconds))
    finish = dt.date.today()
    days = (finish - start).days
    if days > 14 and not force:
        raise Exception('Too many days (%d) - ' % days +
                        'do a bulk download instead: https://www.garmin.com/en-US/account/datamanagement/')
    if days:
        while start < finish:
            yield start
            start += dt.timedelta(days=1)
    else:
        log.warning('No dates to download')


class MonitorLoaderMixin:

    def _resolve_duplicate(self, name, instance, prev):
        log.warning(f'Using max of duplicate values at {instance.time} for {name} '
                    f'({instance.value}/{prev.value})')
        prev.value = max(prev.value, instance.value)


class SqliteMonitorLoader(MonitorLoaderMixin, SqliteLoader): pass


class PostgresqlMonitorLoader(MonitorLoaderMixin, PostgresqlLoader): pass



NEW_STEPS = N._new(N.STEPS)
STEPS_DESCRIPTION = '''The increment in steps read from the FIT file.'''


class MonitorReader(MultiProcFitReader, LoaderMixin):

    '''
    These overlap (well, often one starts when another ends),
    so we read everything in and then, at the end, remove overlaps.
    '''

    loaders = {SQLITE: SqliteMonitorLoader,
               POSTGRESQL: PostgresqlMonitorLoader}

    def __init__(self, *args, **kargs):
        from ...commands.read import MONITOR
        super().__init__(*args, sub_dir=MONITOR, **kargs)

    def _get_loader(self, s, **kargs):
        if 'owner' not in kargs:
            kargs['owner'] = self.owner_out
        return super()._get_loader(s, **kargs)

    def _base_command(self):
        force = mm(FORCE) if self.force else ""
        return f'{READ} {force}'

    @staticmethod
    def parse_records(data):
        return MonitorReader.read_fit_file(data, merge_duplicates, fix_degrees, unpack_single_bytes)

    @staticmethod
    def read_first_timestamp(path, records):
        return MonitorReader._first(path, records, MONITORING_INFO_ATTR).value.timestamp

    @staticmethod
    def read_last_timestamp(path, records):
        return MonitorReader._last(path, records, MONITORING_ATTR).value.timestamp

    def _read_data(self, s, file_scan):
        records = self.parse_records(read_fit(file_scan.path))
        first_timestamp = self.read_first_timestamp(file_scan.path, records)
        last_timestamp = self.read_last_timestamp(file_scan.path, records)
        if first_timestamp == last_timestamp:
            log.warning('File %s is empty (no timespan)' % file_scan)
            raise AbortImportButMarkScanned()
        if not first_timestamp:
            raise Exception('Missing timestamp in %s' % file_scan)

        log.info(f'Importing monitor data from {file_scan} '
                 f'for {format_time(first_timestamp)} - {format_time(last_timestamp)}')
        if self.force:
            log.debug(f'Deleting previous entry')
            s.query(MonitorJournal).filter(MonitorJournal.file_hash == file_scan.file_hash).delete()
        else:
            if s.query(MonitorJournal).filter(MonitorJournal.file_hash == file_scan.file_hash).count():
                raise Exception(f'Duplicate for {file_scan.path}')  # should never happen
        # adding 0.1s to the end time makes the intervals semi-open which simplifies cleanup later
        mjournal = add(s, MonitorJournal(start=first_timestamp,
                                         finish=last_timestamp + dt.timedelta(seconds=0.1),
                                         file_hash_id=file_scan.file_hash.id))
        return mjournal, (first_timestamp, last_timestamp, mjournal, records)

    def _load_data(self, s, loader, data):
        first_timestamp, last_timestamp, mjournal, records = data
        steps_by_activity = defaultdict(lambda: 0)
        for record in records:
            if HEART_RATE_ATTR in record.data and record.data[HEART_RATE_ATTR][0][0]:
                loader.add(T.HEART_RATE, Units.BPM, None, mjournal,
                           record.data[HEART_RATE_ATTR][0][0], record.timestamp, StatisticJournalInteger,
                           description='''The instantaneous heart rate.''')
            if STEPS_ATTR in record.data:
                # we ignore activity type here (used to store it when activity group and statistic name
                # were mixed together, but never used it anywhere)
                reset = False
                for activity, steps in zip(record.data[ACTIVITY_TYPE_ATTR][0], record.data[STEPS_ATTR][0]):
                    reset = reset or steps < steps_by_activity[activity]
                if reset: steps_by_activity = defaultdict(lambda: 0)
                for activity, steps in zip(record.data[ACTIVITY_TYPE_ATTR][0], record.data[STEPS_ATTR][0]):
                    steps_by_activity[activity] = steps
                total = sum(steps_by_activity.values())
                loader.add(T.CUMULATIVE_STEPS, Units.STEPS_UNITS, None,
                           mjournal, total,
                           record.timestamp, StatisticJournalInteger,
                           description='''The number of steps in a day to this point in time.''')

    def _shutdown(self, s):
        super()._shutdown(s)
        if not self.worker:
            log.info('Calculating differential in main thread')
            self._fix_overlapping_monitors(s)
            self._update_differential(s)

    def _fix_overlapping_monitors(self, s):
        pair = self._next_overlap(s)
        while pair:
            self._fix_pair(s, *pair)
            pair = self._next_overlap(s)
        s.commit()

    def _fix_pair(self, s, a, b):
        # a starts before b (from query)
        if b.finish <= a.finish:
            # b completely enclosed in a
            log.warning(f'Deleting monitor journal entry that completely overlaps another')
            log.debug(f'{a.start} - {a.finish} ({a.id}) encloses {b.start} - {b.finish} ({b.id})')
            # be careful to delete superclass...
            s.query(Source).filter(Source.id == b.id).delete()
        else:
            # otherwise, shorten a so it finishes where b starts
            q = s.query(StatisticJournal). \
                filter(StatisticJournal.source == a,
                       StatisticJournal.time >= b.start)
            count = q.count()
            if count:
                # not really a warning because we expect this
                log.debug(f'Shifting edge of overlapping monitor journals ({count} statistic values)')
                log.debug(f'{a.start} - {a.finish} ({a.id}) overlaps {b.start} - {b.finish} ({b.id})')
                q.delete()
            # update monitor whether statistics were changed or not
            log.debug(f'Shift monitor finish back from {a.finish} to {b.start}')
            a.finish = b.start
            s.flush()  # not sure this is needed

    def _next_overlap(self, s):
        MonitorJournal2 = aliased(MonitorJournal)
        row = s.query(MonitorJournal, MonitorJournal2). \
                  select_from(MonitorJournal). \
                  join(MonitorJournal2,
                       # two overlaps, order, and not self and not null
                       and_(MonitorJournal2.finish >= MonitorJournal.start,
                            MonitorJournal2.start < MonitorJournal.finish,
                            MonitorJournal.start <= MonitorJournal2.start,
                            MonitorJournal.id != MonitorJournal2.id)). \
                  order_by(MonitorJournal.start).first()
        return row

    def _update_differential(self, s):
        # this reads CUMULATIVE_STEPS (which is what was in the files) and any existing STEPS
        # then calculates what STEPS should be and fixes up any incorrect or missing data
        # (i guess maybe slightly faster when running incrementally)
        df = self._read_diff(s)
        df = self._calculate_diff(df)
        self._write_diff(s, df)

    def _read_diff(self, s):
        qs = s.query(StatisticJournalInteger.time.label(N.TIME),
                     StatisticJournalInteger.value.label(N.STEPS)). \
            join(StatisticName). \
            filter(StatisticName.name == N.STEPS,
                   StatisticName.owner == self.owner_out).cte()
        q = s.query(StatisticJournalInteger.time.label(N.TIME),
                    StatisticJournalInteger.source_id.label(N.SOURCE),
                    StatisticJournalInteger.value.label(N.CUMULATIVE_STEPS),
                    qs.c.steps.label(N.STEPS)). \
            join(StatisticName).outerjoin(qs, StatisticJournalInteger.time == qs.c.time). \
            filter(StatisticName.name == N.CUMULATIVE_STEPS,
                   StatisticName.owner == self.owner_out). \
            order_by(StatisticJournalInteger.time)
        # log.debug(q)
        df = read_query(q, index=N.TIME)
        return df

    def _calculate_diff(self, df):
        df[NEW_STEPS] = df[N.CUMULATIVE_STEPS].diff()
        df.loc[df[NEW_STEPS] < 0, NEW_STEPS] = df[N.CUMULATIVE_STEPS]
        df.loc[df[NEW_STEPS].isna(), NEW_STEPS] = df[N.CUMULATIVE_STEPS]
        return df

    def _write_diff(self, s, df):
        steps = StatisticName.add_if_missing(s, T.STEPS, StatisticJournalType.INTEGER, Units.STEPS_UNITS,
                                             None, self.owner_out, description=STEPS_DESCRIPTION)
        times = df.loc[(df[NEW_STEPS] != df[N.STEPS]) & ~df[N.STEPS].isna()].index.astype(np.int64) / 1e9
        if len(times):
            n = s.query(func.count(StatisticJournal.id)). \
                filter(StatisticJournal.time.in_(times),
                       StatisticJournal.statistic_name == steps).scalar()
            log.warning(f'Deleting {n} {N.STEPS} entries')
            s.query(StatisticJournal.id). \
                filter(StatisticJournal.time.in_(times),
                       StatisticJournal.statistic_name == steps).delete(synchronize_session=False)
        loader = self._get_loader(s, owner=self.owner_out, add_serial=False)
        for time, row in df.loc[(df[NEW_STEPS] != df[N.STEPS]) & ~df[NEW_STEPS].isna()].iterrows():
            loader.add(T.STEPS, Units.STEPS_UNITS, None, row[N.SOURCE], int(row[NEW_STEPS]),
                       time, StatisticJournalInteger, description=STEPS_DESCRIPTION)
        loader.load()




import numpy as np
import pandas as pd

from .heart_rate import BC_ZONES
from ..lib.data import interpolate_freq, safe_none, safe_return
from ..lib.date import to_time
from ..names import Names, Titles


@safe_none
def hr_zone(heart_rate_df, fthr_df, pc_fthr_zones=BC_ZONES, heart_rate=Names.HEART_RATE, hr_zone=Names.HR_ZONE):
    '''
    mutate input df to include hr zone

    this can be used for a huge slew of data, or for an individual activity.
    '''
    fthrs = sorted([(time, row[Names.FTHR]) for time, row in fthr_df.dropna().iterrows()], reverse=True)
    if not fthrs:
        raise Exception(f'No {Names.FTHR} defined')
    fthrs = fthrs + [(to_time('2100'), None)]
    fthrs = [(a[0], b[0], a[1]) for a, b in zip(fthrs, fthrs[1:])]
    heart_rate_df[hr_zone] = np.nan
    for start, finish, fthr in fthrs:  # start is inclusive
        start, finish = pd.to_datetime(start, utc=True), pd.to_datetime(finish, utc=True)
        zones = [x * fthr / 100.0 for x in pc_fthr_zones]
        for zone, upper in enumerate(zones + [None], start=1):
            if zone == 1:
                # we don't try to distinguish zones below 1
                heart_rate_df.loc[(heart_rate_df.index >= start) & (heart_rate_df.index < finish) &
                                  (heart_rate_df[heart_rate] <= upper),
                                  [hr_zone]] = zone
            elif not upper:
                # or above 5
                heart_rate_df.loc[(heart_rate_df.index >= start) & (heart_rate_df.index < finish) &
                                  (heart_rate_df[heart_rate] > lower),
                                  [hr_zone]] = zone
            else:
                hrz = ((heart_rate_df.loc[(heart_rate_df.index >= start) & (heart_rate_df.index < finish) &
                                          (heart_rate_df[heart_rate] > lower) & (heart_rate_df[heart_rate] <= upper),
                                          [heart_rate]] - lower) / (upper - lower)) + zone
                # .values below from
                # https://stackoverflow.com/questions/12307099/modifying-a-subset-of-rows-in-a-pandas-dataframe
                # i do not understand why it is needed...
                heart_rate_df.loc[(heart_rate_df.index >= start) & (heart_rate_df.index < finish) &
                                  (heart_rate_df[heart_rate] > lower) & (heart_rate_df[heart_rate] <= upper),
                                  [hr_zone]] = hrz.values
            lower = upper


@safe_return(lambda: pd.DataFrame(columns=[Names.HR_IMPULSE_10]))
def impulse_10(hr_zone_df, impulse, hr_zone=Names.HR_ZONE):
    '''
    interpolate HR to 10s values then calculate impulse using model parameters.

    this can be used for a huge slew of data, or for an individual activity.
    '''
    impulse_df = interpolate_freq(hr_zone_df.loc[:, [hr_zone]], '10s',
                                  method='index', limit=int(0.5 + impulse.max_secs / 10)).dropna()
    impulse_df[Names.HR_IMPULSE_10] = (impulse_df[hr_zone] - impulse.zero) / (impulse.one - impulse.zero)
    impulse_df[Names.HR_IMPULSE_10].clip(lower=0, inplace=True)
    impulse_df[Names.HR_IMPULSE_10] = impulse_df[Names.HR_IMPULSE_10] ** impulse.gamma
    impulse_df.drop(columns=[hr_zone], inplace=True)
    return impulse_df




from bokeh.models import PanTool, ZoomInTool, ZoomOutTool, ResetTool, HoverTool, LinearAxis
from bokeh.plotting import figure

from .line import multi_dot_plot, dot_plotter, comb_plotter, DEFAULT_BACKEND
from .utils import make_range, evenly_spaced_hues, tooltip
from ..frame import related_statistics
from ...names import Names as N, like, Units


def std_distance_time_plot(nx, ny, source, x_range=None, output_backend=DEFAULT_BACKEND):
    # avoid range errors
    if len(source[N.ACTIVE_TIME_S].dropna()) < 2:
        return None
    groups = [group for statistic, group in related_statistics(source, N.ACTIVE_TIME)]
    if not groups:
        # original monochrome plot
        return multi_dot_plot(nx, ny, N.TIME, [N.ACTIVE_TIME_H, N.ACTIVE_DISTANCE_KM], source,
                              ['black', 'grey'], alphas=[1, 0.5], x_range=x_range, rescale=True)
    times = [f'{N.ACTIVE_TIME}:{group}' for group in groups]
    distances = [f'{N.ACTIVE_DISTANCE}:{group}' for group in groups]
    time_y_range = make_range(source[N.ACTIVE_TIME_H])
    distance_y_range = make_range(source[N.ACTIVE_DISTANCE_KM])
    colours = list(evenly_spaced_hues(len(groups)))
    tooltip_names = [N.ACTIVE_TIME_H, N.ACTIVE_DISTANCE_KM, N.ACTIVITY_GROUP, N.LOCAL_TIME]
    tooltip_names += [name for name in like(N._delta(N.FITNESS_ANY), source.columns) if ':' not in name]
    tooltip_names += [name for name in like(N._delta(N.FATIGUE_ANY), source.columns) if ':' not in name]
    tools = [PanTool(dimensions='width'),
             ZoomInTool(dimensions='width'), ZoomOutTool(dimensions='width'),
             ResetTool(),
             HoverTool(tooltips=[tooltip(name) for name in tooltip_names], names=['with_hover'])]
    f = figure(output_backend=output_backend, plot_width=nx, plot_height=ny, x_axis_type='datetime', tools=tools)
    f.yaxis.axis_label = f'lines - {N.ACTIVE_TIME_H}'
    f.y_range = time_y_range
    f.extra_y_ranges = {N.ACTIVE_DISTANCE: distance_y_range}
    f.add_layout(LinearAxis(y_range_name=N.ACTIVE_DISTANCE, axis_label=f'dots - {N.ACTIVE_DISTANCE_KM}'), 'right')
    plotter = comb_plotter()
    for time, colour, group in zip(times, colours, groups):
        time_h = N._slash(time, Units.H)
        source[time_h] = source[time] / 3600
        source[N.ACTIVITY_GROUP] = group
        plotter(f, x=N.TIME, y=time_h, source=source, color=colour, alpha=1)
    plotter = dot_plotter()
    for distance, colour, group in zip(distances, colours, groups):
        distance_km = N._slash(distance, Units.KM)
        source[distance_km] = source[distance]
        source[N.ACTIVITY_GROUP] = group
        plotter(f, x=N.TIME, y=distance_km, source=source, color=colour, alpha=1, name='with_hover',
                y_range_name=N.ACTIVE_DISTANCE)
    f.xaxis.axis_label = N.TIME
    f.toolbar.logo = None
    if ny < 300: f.toolbar_location = None
    if x_range: f.x_range = x_range
    return f




from logging import getLogger
from subprocess import run
from tempfile import TemporaryDirectory

import sqlalchemy.sql.functions as func

from ch2.commands.read import read
from ch2.commands.args import bootstrap_dir, m, V, DEV, mm, BASE, MONITOR
from ch2.config.profile.default import default
from ch2.lib.date import to_time, local_date_to_time
from ch2.sql.tables.monitor import MonitorJournal
from ch2.sql.tables.pipeline import PipelineType
from ch2.sql.tables.statistic import StatisticJournal, StatisticName
from ch2.pipeline.calculate.monitor import MonitorCalculator
from ch2.pipeline.pipeline import run_pipeline
from ch2.data import Names as N
from tests import LogTestCase

log = getLogger(__name__)


class TestMonitor(LogTestCase):

    def test_monitor(self):
        with TemporaryDirectory() as f:
            args, data = bootstrap_dir(f, m(V), '5')
            bootstrap_dir(f, m(V), '5', mm(DEV), configurator=default)
            args, data = bootstrap_dir(f, m(V), '5', mm(DEV),
                                       'read', 'data/test/source/personal/25822184777.fit')
            read(args, data)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            run_pipeline(data, PipelineType.CALCULATE, force=True, start='2018-01-01', n_cpu=1)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            with data.db.session_context() as s:
                n = s.query(func.count(StatisticJournal.id)).scalar()
                self.assertEqual(n, 137)
                mjournal = s.query(MonitorJournal).one()
                self.assertNotEqual(mjournal.start, mjournal.finish)

    def test_values(self):
        with TemporaryDirectory() as f:
            bootstrap_dir(f, m(V), '5')
            bootstrap_dir(f, m(V), '5', mm(DEV), configurator=default)
            for file in ('24696157869', '24696160481', '24696163486'):
                args, data = bootstrap_dir(f, m(V), '5', mm(DEV),
                                           'read', mm(MONITOR),
                                           'data/test/source/personal/andrew@acooke.org_%s.fit' % file)
                read(args, data)
            # path = args.system_path(subdir='data', file='activity.db')
            # run(f'sqlite3 {path} ".dump"', shell=True)
            run_pipeline(data, PipelineType.CALCULATE, force=True, like=('%Monitor%',), start='2018-01-01', n_cpu=1)
            with data.db.session_context() as s:
                mjournals = s.query(MonitorJournal).order_by(MonitorJournal.start).all()
                assert mjournals[2].start == to_time('2018-09-06 15:06:00'), mjournals[2].start
                # steps
                summary = s.query(StatisticJournal).join(StatisticName). \
                    filter(StatisticJournal.time >= local_date_to_time('2018-09-06'),
                           StatisticJournal.time < local_date_to_time('2018-09-07'),
                           StatisticName.owner == MonitorCalculator,
                           StatisticName.name == N.DAILY_STEPS).one()
                if summary.value != 12757:
                    path = args.system_path(subdir='data', file='activity.db')
                    run('sqlite3 %s "select * from statistic_journal as j, statistic_journal_integer as i, '
                        'statistic_name as n where j.id = i.id and j.statistic_name_id = n.id and '
                        'n.name = \'steps\' order by j.time"' % path, shell=True)
                    run('sqlite3 %s "select * from statistic_journal as j, statistic_journal_integer as i, '
                        'statistic_name as n where j.id = i.id and j.statistic_name_id = n.id and '
                        'n.name = \'cumulative-steps\' order by j.time"' % path, shell=True)
                # connect has 12757 for this date,
                self.assertEqual(summary.value, 12757)

    FILES = ('25505915679', '25519562859', '25519565531', '25532154264', '25539076032', '25542112328')

    def generic_bug(self, files, join=False):
        with TemporaryDirectory() as f:
            args, data = bootstrap_dir(f, m(V), '5')
            bootstrap_dir(f, m(V), '5', mm(DEV), configurator=default)
            if join:
                files = ['data/test/source/personal/andrew@acooke.org_%s.fit' % file for file in files]
                args, data = bootstrap_dir(f, mm(DEV), 'read', *files)
                read(args, data)
            else:
                for file in files:
                    args, data = bootstrap_dir(f, mm(DEV), 'read',
                                               'data/test/source/personal/andrew@acooke.org_%s.fit' % file)
                    read(args, data)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            with data.db.session_context() as s:
                # steps
                summary = s.query(StatisticJournal).join(StatisticName). \
                    filter(StatisticJournal.time >= local_date_to_time('2018-10-07'),
                           StatisticJournal.time < local_date_to_time('2018-10-08'),
                           StatisticName.owner == MonitorCalculator,
                           StatisticName.name == N.DAILY_STEPS).one()
                # connect has 3031 for this date.
                self.assertEqual(summary.value, 3031)

    def test_bug(self):
        self.generic_bug(sorted(self.FILES))

    def test_bug_reversed(self):
        self.generic_bug(sorted(self.FILES, reverse=True))

    def test_bug_join(self):
        self.generic_bug(sorted(self.FILES), join=True)

    def test_bug_reversed_join(self):
        self.generic_bug(sorted(self.FILES, reverse=True), join=True)

    # issue 6
    def test_empty_data(self):
        with TemporaryDirectory() as f:
            args, data = bootstrap_dir(f, m(V), '5')
            bootstrap_dir(f, m(V), '5', mm(DEV), configurator=default)
            args, data = bootstrap_dir(f, m(V), '5', mm(DEV),
                                      'read', 'data/test/source/other/37140810636.fit')
            read(args, data)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            run_pipeline(data, PipelineType.CALCULATE, n_cpu=1)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            with data.db.session_context() as s:
                n = s.query(func.count(StatisticJournal.id)).scalar()
                self.assertEqual(n, 44)
                mjournal = s.query(MonitorJournal).one()
                self.assertNotEqual(mjournal.start, mjournal.finish)




from sqlalchemy.ext.compiler import compiles
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.sql import expression, case
from sqlalchemy.types import Numeric

Base = declarative_base()
SystemBase = declarative_base()


# https://docs.sqlalchemy.org/en/13/core/compiler.html


class greatest(expression.FunctionElement):
    type = Numeric()
    name = 'greatest'


@compiles(greatest)
def default_greatest(element, compiler, **kw):
    return compiler.visit_function(element)


@compiles(greatest, 'sqlite')
@compiles(greatest, 'mssql')
@compiles(greatest, 'oracle')
def case_greatest(element, compiler, **kw):
    arg1, arg2 = list(element.clauses)
    return compiler.process(case([(arg1 > arg2, arg1)], else_=arg2), **kw)




from logging import getLogger

from .utils import MultiProcCalculator, ActivityJournalCalculatorMixin, DataFrameCalculatorMixin
from ...data import Statistics
from ...data.elevation import smooth_elevation
from ...data.frame import present, valid
from ...names import N, Titles, Units
from ...sql import StatisticJournalFloat

log = getLogger(__name__)


class ElevationCalculator(ActivityJournalCalculatorMixin, DataFrameCalculatorMixin, MultiProcCalculator):

    def __init__(self, *args, smooth=3, **kargs):
        self.smooth = smooth
        super().__init__(*args, **kargs)

    def _read_dataframe(self, s, ajournal):
        from ..owners import SegmentReader
        try:
            return Statistics(s, activity_journal=ajournal, with_timespan=True). \
                by_name(SegmentReader, N.DISTANCE, N.RAW_ELEVATION, N.ELEVATION, N.ALTITUDE).df
        except Exception as e:
            log.warning(f'Failed to generate statistics for elevation: {e}')
            raise

    def _calculate_stats(self, s, ajournal, df):
        if not present(df, N.ELEVATION):
            if present(df, N.RAW_ELEVATION):
                df = smooth_elevation(df, smooth=self.smooth)
            elif present(df, N.ALTITUDE):
                log.warning(f'Using {N.ALTITUDE} as {N.ELEVATION}')
                df[N.ELEVATION] = df[N.ALTITUDE]
            return df
        else:
            return None

    def _copy_results(self, s, ajournal, loader, df):
        for time, row in df.iterrows():
            if N.ELEVATION in row and valid(row[N.ELEVATION]):
                loader.add(Titles.ELEVATION, Units.M, None, ajournal, row[N.ELEVATION],
                           time, StatisticJournalFloat,
                           description='An estimate of elevation (may come from various sources).')
            if N.GRADE in row and valid(row[N.GRADE]):
                loader.add(Titles.GRADE, Units.PC, None, ajournal, row[N.GRADE],
                           time, StatisticJournalFloat,
                           description='The gradient of the smoothed SRTM1 elevation.')
